#!/usr/bin/env python3
"""
This file contains JSON-based templates for EMSO metadata

Conventions:

  '*' Attributes are mandatory, e.g. "*long_name"
  '~' Attributes are optional, if no provided they will be automatically derived from other info
  '$' Attributes are mandatory, but if not present the generator will ask the user to choose it from a list

author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 19/4/23
"""

import rich
import time


def variable_metadata():
    """
    Returns all the metadata required in a variable
    """
    return {
        "*long_name": "",
        "*sdn_parameter_uri": "",
        "~sdn_uom_uri": "",
        "~standard_name": "",
    }


def sensor_metadata():
    """
    Returns all the metadata required for a sensor
    """
    return {
        "*sensor_model_uri": "",
        "*sensor_serial_number": "",
        "$sensor_mount": "",
        "$sensor_orientation": ""
    }


def quality_control_metadata(long_name):
    """
    Returns the minimal attributes for a quality control variable
    """
    return {
        "long_name": long_name + " quality control flags",
        "conventions": "OceanSITES QC Flags",
        "flag_values": [0, 1, 2, 3, 4, 7, 8, 9],
        "flag_meanings": ["unknown", "good_data", "probably_good_data", "potentially_correctable_bad_data",
                          "bad_data", "nominal_value", "interpolated_value", "missing_value"]
    }

def dimension_metadata(dim):

    __dimension_metadata = {
        "TIME": {
            "long_name": "time of measurements",
            # UNIX time: seconds since 1970
            "sdn_parameter_uri": "https://vocab.nerc.ac.uk/collection/P01/current/ELTMEP01/",
            "standard_name": "time",
            # Days since 1950
            # "sdn_parameter_uri": "https://vocab.nerc.ac.uk/collection/P01/current/ELTJLD01/"
            "axis": "T",
            "sdn_uom_uri": "http://vocab.nerc.ac.uk/collection/P06/current/UTBB/"
        },
        "DEPTH": {
            "long_name": "depth of measurements",
            "sdn_parameter_uri": "https://vocab.nerc.ac.uk/collection/P01/current/ADEPZZ01",
            "standard_name": "depth",
            "axis": "Z",
            "sdn_uom_uri": "http://vocab.nerc.ac.uk/collection/P06/current/ULAA/"
        },
        "LATITUDE": {
            "long_name": "latitude of measurements",
            "sdn_parameter_uri": "https://vocab.nerc.ac.uk/collection/P01/current/ALATZZ01",
            "standard_name": "latitude",
            "axis": "Y",
            "sdn_uom_uri": "http://vocab.nerc.ac.uk/collection/P06/current/UAAA/"
        },
        "LONGITUDE": {
            "long_name": "longitude of measurements",
            "sdn_parameter_uri": "https://vocab.nerc.ac.uk/collection/P01/current/ALONZZ01",
            "standard_name": "longitude",
            "axis": "X",
            "sdn_uom_uri": "http://vocab.nerc.ac.uk/collection/P06/current/UAAA/"
        },
        "SENSOR_ID": {
            "long_name": "Identifier of the sensor that took the measurement"
        }
    }
    if dim not in __dimension_metadata.keys():
        raise LookupError(f"Not a valid dimension '{dim}'. Expected one of the following {list(__dimension_metadata.keys())}")
    return __dimension_metadata[dim].copy()


def global_metadata():
    """
    Returns all the metadata required for a sensor
    """
    return {
        "*title": "",
        "*summary": "",
        "~Conventions": "",
        "*institution_edmo_code": "",
        "~update_interval": "",
        "$site_code": "",
        "$emso_facility": "",
        "*source": "",
        "$data_type": "",
        "~format_version": "",
        "~network": "",
        "$data_mode": "",
        "project": "",
        "*principal_investigator": "",
        "*principal_investigator_email": "",
        "~license": ""
    }


def user_selectable_attributes():
    """
    Returns a dict with all the user-selectable metadata attributes
    """
    return {
        "sensor_mount": [
            "mounted_on_fixed_structure",
            "mounted_on_surface_buoy",
            "mounted_on_mooring_line",
            "mounted_on_bottom_lander",
            "mounted_on_moored_profiler",
            "mounted_on_glider",
            "mounted_on_shipborne_fixed",
            "mounted_on_shipborne_profiler",
            "mounted_on_seafloor_structure",
            "mounted_on_benthic_node",
            "mounted_on_benthic_crawler",
            "mounted_on_surface_buoy_tether",
            "mounted_on_seafloor_structure_riser",
            "mounted_on_fixed_subsurface_vertical_profile"
        ],
        "sensor_orientation": ["downward", "upward", "horizontal"],
        "data_type": ["OceanSITES profile data", "OceanSITES time-series data", "OceanSITES trajectory data"],
        "emso_facility": [
            "Azores",
            "Black Sea",
            "Canary Islands",
            "Cretan Sea",
            "Hellenic Arc",
            "Iberian Margin",
            "Western Ligurian Sea",
            "Eastern Ligurian Sea",
            "Molene",
            "OBSEA",
            "SmartBay",
            "South Adriatic Sea",
            "Western Ionian Sea",
            "Western Mediterranean Sea"
        ],
        "site_code": [
            "BB",
            "FF",
            "ALBATROSS-MII (seafloor)",
            "ALBATROSS-MII (mooring)",
            "SJB",
            "DYFAMED",
            "EMSO NICE",
            "W1M3A",
            "OBSEA (seafloor)",
            "OBSEA (buoy)",
            "NEMO-SN1",
            "E2M3A",
            "PYLOS",
            "E1M3A",
            "IbMa-CSV",
            "ESTOC - fixed station",
            "ESTOC - gliders",
            "BOREL2",
            "SeaMoN West",
            "SeaMoN East"
        ],
        "data_mode": [
            "R --> Real-time data", "P --> Provisional data", "D --> Delayed-mode data", "M --> Mixed"
        ]
    }


def choose_interactively(attr_name: str, hint: str, options: list) -> str:
    """
    Asks the user to choose an option from within a list in an interactive manner
    """
    valid_option = False
    n = len(options)
    while not valid_option:
        rich.print(f"[cyan]Select one of the following values for the '{attr_name}' attribute ('{hint}')")
        for i in range(n):
            if i < 9:
                rich.print(" ", end="")
            rich.print(f"{i + 1} - {options[i]}")
        inp = input("Selection: ")
        try:
            selection = int(inp.strip())
        except ValueError:  # invalid user input
            selection = -1

        if type(selection) == int and selection > 0 and selection <= n:
            valid_option = True
        else:
            rich.print("[red]User input not valid! try again...")
            time.sleep(1)
    result = options[selection - 1]
    if "--> " in result:
        result = result.split(" --> ")[0]
    rich.print(f"User selection [purple]'{result}'[/purple]")
    return result


import pandas as pd
import rich

from .autofill import autofill_waterframe_coverage
from .constants import iso_time_format
from .dataset import consolidate_metadata, set_multisensor
from .utils import merge_dicts
from .waterframe import WaterFrame


def merge_waterframes(waterframes):
    """
    Combine all WaterFrames into a single waterframe. Both data and metadata are consolidated into a single
    structure
    """
    dataframes = []  # list of dataframes
    global_attr = []  # list of dict containing global attributes
    variables_attr = {}  # dict all the variables metadata
    for wf in waterframes:
        df = wf.data
        # setting time as the index
        df = df.set_index("TIME")
        df = df.drop_duplicates(keep="first")
        df = df.sort_index(ascending=True)
        df["SENSOR_ID"] = wf.metadata["$sensor_id"]

        dataframes.append(df)
        global_attr.append(wf.metadata)
        for varname, varmeta in wf.vocabulary.items():
            if varname not in variables_attr.keys():
                variables_attr[varname] = [varmeta]  # list of dicts with metadata
            else:
                variables_attr[varname].append(varmeta)

    df = pd.concat(dataframes)  # Consolidate data in a single dataframe
    df = df.sort_index(ascending=True)  # sort by date
    df = df.reset_index()  # get back to numerical index

    # Consolidating Global metadata, the position in the array is the priority
    global_meta = {}
    for g in reversed(global_attr):  # loop backwards, last element has lower priority
        global_meta = merge_dicts(g, global_meta)

    variable_meta = {}
    for varname, varmeta in variables_attr.items():
        variable_meta[varname] = consolidate_metadata(varmeta)

    wf = WaterFrame(df, global_meta, variable_meta)

    try:
        wf = set_multisensor(wf)
    except LookupError as e:
        rich.print(f"[red]ERROR {e}")

    wf = autofill_waterframe_coverage(wf)  # update the coordinates max/min in metadata

    # Add versioning info
    now = pd.Timestamp.now(tz="utc").strftime(iso_time_format)
    if len(waterframes) > 1:
        # New waterframe
        wf.metadata["date_created"] = now
        wf.metadata["date_modified"] = now
    else:  # just update the date_modified
        if "date_created" not in wf.metadata.keys():
            wf.metadata["date_created"] = now
        wf.metadata["date_modified"] = wf.metadata["date_created"]
    return wf



#!/usr/bin/env python3
"""
This python project connects to an ERDDAP service and ensures that all listed datasets are compliant with the EMSO
Harmonization Guidelines.

author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 23/2/23
"""
import json
import os
import rich
import time
from .erddap import ERDDAP
import pandas as pd
from .metadata import EmsoMetadata
from .metadata.utils import threadify
from .metadata.dataset import get_netcdf_metadata
from .metadata.tests import EmsoMetadataTester


def metadata_report(target,
                    datasets: list=[],
                    just_list: bool=False,
                    just_print: bool=False,
                    verbose: bool = False,
                    save_metadata: str = "",
                    output: str = "",
                    report: bool = False,
                    clear: bool = False,
                    excel_table: bool = False
                    ):
    """

    :param target: ERDDAP service URL, NetCDF file or JSON metadata file
    :param datasets: List of datasets to check (by default check all of them)
    :param just_list: Just list the available datasets and exit
    :param just_print: Just pretty-print the dataset metadata
    :param verbose: Shows more info
    :param save_metadata: Save dataset's metadata into the specified folder
    :param output: file to store the report of all the datasets
    :param report: Gnerate a CSV file for every test
    :param clear: Clears the downloaded files
    param: excel_table:  prints the results in a excel compatible table
    """
    if clear:
        rich.print("Clearing downloaded files...", end="")
        EmsoMetadata.clear_downloads()
        rich.print("[green]done")
        exit()

    if not target:
        rich.print("[red]ERDDAP URL, NetCDF file or JSON file required!")
        exit()

    if target.startswith("http"):
        # Assuming ERDDAP service
        erddap = ERDDAP(target)
        datasets = datasets
        if not datasets:  # If a list of datasets is not provided, use all datasets in the service
            datasets = erddap.dataset_list()

        if just_list:  # If set, just list datasets and exit
            datasets = erddap.dataset_list()
            rich.print("[green]Listing datasets in ERDDAP:")
            for i in range(len(datasets)):
                rich.print(f"    {i:02d} - {datasets[i]}")
            exit()

        # Get all Metadata from all datasets
        t = time.time()
        tasks = [(dataset_id,) for dataset_id in datasets]
        datasets_metadata = threadify(tasks, erddap.dataset_metadata, max_threads=5)
        rich.print(f"Getting metadata from ERDDDAP took {time.time() - t:.02f} seconds")

    # Processing NetCDF file
    elif target.endswith(".nc"):
        rich.print(f"Loading metadata from file {target}")
        metadata = get_netcdf_metadata(target)
        datasets_metadata = [metadata]

    # Processing JSON file
    elif target.endswith(".json"):
        rich.print(f"Loading metadata from file {target}")
        with open(target) as f:
            metadata = json.load(f)
        datasets_metadata = [metadata]  # an array with only one value
    else:
        rich.print("[red]Invalid arguments! Expected an ERDDAP url, a NetCDF file or a JSON file")

    if just_print:
        for d in datasets_metadata:
            rich.print(d)
            exit()

    if save_metadata:
        os.makedirs(save_metadata, exist_ok=True)
        rich.print(f"Saving datasets metadata in '{save_metadata}' folder")
        for dataset_id in datasets:
            file = os.path.join(save_metadata, f"{dataset_id}.json")
            metadata = erddap.dataset_metadata(dataset_id)
            with open(file, "w") as f:
                f.write(json.dumps(metadata, indent=2))
        exit()

    tests = EmsoMetadataTester()

    total = []
    required = []
    optional = []
    institution = []
    emso_facility = []
    dataset_id = []
    for i in range(len(datasets_metadata)):
        metadata = datasets_metadata[i]
        r = tests.validate_dataset(metadata, verbose=verbose, store_results=report)
        total.append(r["total"])
        required.append(r["required"])
        optional.append(r["optional"])
        institution.append(r["institution"])
        emso_facility.append(r["emso_facility"])
        dataset_id.append(r["dataset_id"])

    tests = pd.DataFrame(
        {
            "dataset_id": dataset_id,
            "emso_facility": emso_facility,
            "institution": institution,
            "total": total,
            "required": required,
            "optional": optional,
        })

    if output:
        rich.print(f"Storing tests results in {output}...", end="")
        tests.to_csv(output, index=False, sep="\t")
        rich.print("[green]done")





#!/usr/bin/env python3
"""
Automatically generates datasets.xml configuration to serve NetCDFs file through ERDDAP


author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 15/5/23
"""
from .erddap.datasets_xml import generate_erddap_dataset, add_dataset
import rich

from .metadata.dataset import load_data


def erddap_config(file: str, dataset_id: str, source_path: str, output: str = "", datasets_xml_file: str = ""):
    """
    :param file: NetCDF file to be used for the configuration
    :param dataset_id: Dataset ID in ERDDAP
    :param source_path: path where the files for this dataset will be stored
    :param output: If set, the dataset configuration will be stored in a new XML file
    :param datasets_xml: Path to the datasets.xml file. If set the configuration will be appended.
    """
    wf = load_data(file)
    xml_chunk = generate_erddap_dataset(wf, source_path, dataset_id=dataset_id)

    if output:
        with open(output, "w") as f:
            f.write(xml_chunk)

    if datasets_xml_file:
        add_dataset(datasets_xml_file, xml_chunk)

    if not datasets_xml_file and not output:
        rich.print(xml_chunk)



#!/usr/bin/env python3
"""

Generates NetCDF files based on CSV files and input from the user

author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 13/4/23
"""
import json
import rich
import pandas as pd
from .metadata.autofill import expand_minmeta, autofill_waterframe
from .metadata.dataset import add_coordinates, ensure_coordinates, update_waterframe_metadata, export_to_netcdf, \
    load_data, df_to_wf
from .metadata.merge import merge_waterframes
from .metadata.minmeta import generate_min_meta_template, load_min_meta, load_full_meta, generate_full_metadata
from .metadata import EmsoMetadata
import copy


def generate_metadata(data_files: list, folder):
    """
    Generate the metadata templates for the input file in the target folder
    """
    # If metadata and generate
    for file in data_files:
        rich.print(f"generating minimal metadata template for {file}")
        wf = load_data(file)

        if file.endswith(".csv"):  # For CSV always generate a minimal metdata file
            generate_min_meta_template(wf, folder)
        elif file.endswith(".nc"):
            generate_full_metadata(wf, folder)

    rich.print(f"[green]Please edit the following files and run the generator with the -m option!")


def generate_datasets(data_list: list, metadata_list: list, emso_metadata: EmsoMetadata):
    """
    Merge data fiiles and metadata files into a NetCDF dataset according to EMSO specs. If provided, depths, lats and
    longs will be added to the dataset as dimensions.
    """

    if emso_metadata:
        emso = emso_metadata
    else:
        emso = EmsoMetadata()

    waterframes = []
    for i in range(len(data_list)):
        data = data_list[i]
        metadata = metadata_list[i]

        if type(data) is str:
            wf = load_data(data)
        elif type(data) is pd.DataFrame:
            wf = df_to_wf(data)
        else:
            raise ValueError(f"Data must be a file or DataFrame, got '{type(data)}'")

        if type(metadata) not in [str, dict]:
            raise ValueError(f"Expected str or dict, got '{type(data)}'")

        # If metadata is dict, assume it as minimal
        if type(metadata) is dict:
            minimal_metadata = True
        elif type(metadata) is str and metadata.endswith(".min.json"):
            minimal_metadata = True
            wf.metadata["$minmeta"] = metadata
        elif type(metadata) is str and metadata.endswith(".full.json") or type(metadata):
            minimal_metadata = False
        else:
            raise ValueError("Expected metadata file with extension .full.json or .min.json!")

        if type(metadata) is str:
            with open(metadata) as f:
                metadata = json.load(f)

        # Create deep copy of the metadata
        metadata = copy.deepcopy(metadata)
        if minimal_metadata:
            minmeta = load_min_meta(wf, metadata, emso)

            if "coordinates" in minmeta.keys():
                lat = minmeta["coordinates"]["latitude"]
                lon = minmeta["coordinates"]["longitude"]
                depth = minmeta["coordinates"]["depth"]
                wf = add_coordinates(wf, lat, lon, depth)

            ensure_coordinates(wf)  # make sure that all coordinates are set
            metadata = expand_minmeta(wf, minmeta, emso)

        else:
            rich.print(f"Loading a full metadata file {metadata}...")
            metadata = load_full_meta(wf, metadata)
        wf = update_waterframe_metadata(wf, metadata)
        waterframes.append(wf)
    return waterframes


def generate_dataset(data: list, metadata: list, generate: bool = False, autofill: bool = False, output: str = "",
                     clear: bool = False, emso_metadata=None) -> str:
    wf = None
    if clear:
        rich.print("Clearing downloaded files...", end="")
        EmsoMetadata.clear_downloads()
        rich.print("[green]done")
        exit()

    if generate and metadata:
        raise ValueError("--metadata and --generate cannot be used at the same time!")

    if not generate and not metadata and not autofill:
        raise ValueError("--metadata OR --generate OR --autofill option ust be used!")

    # If metadata and generate
    if generate:
        rich.print("[blue]Generating metadata templates...")
        generate_metadata(data, generate)
        exit()

    if metadata:
        waterframes = generate_datasets(data, metadata, emso_metadata=emso_metadata)

        # If ALL water frames are empty we have nothing else to do, just exit
        some_data = False
        for wf in waterframes:
            if not wf.data.empty:
                some_data = True
        if not some_data:
            rich.print("[red]There is not data in the dataframes! exit")
            exit(0)
        wf = merge_waterframes(waterframes)

    if autofill:
        if len(data) > 1:
            raise ValueError("Only one data file expected!")
        filename = data[0]
        wf = load_data(filename)
        wf = autofill_waterframe(wf)

    if output:
        export_to_netcdf(wf, output)

    if not wf:
        if len(data) > 1:
            raise ValueError("Only one data file expected!")
        filename = data[0]
        wf = load_data(filename)

    if output:
        return output



#!/usr/bin/env python3
"""
This script contains tools to access, download and parse Metadata stored in Markdown format within EMSO ERIC's github
repository.

author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 3/3/23
"""
import os
import ssl
import rich
import pandas as pd
import json
import time
from .utils import download_files, get_file_list

emso_version = "develop"

# List of URLs
emso_metadata_url = f"https://raw.githubusercontent.com/emso-eric/emso-metadata-specifications/{emso_version}/EMSO_metadata.md"
oceansites_codes_url = f"https://raw.githubusercontent.com/emso-eric/emso-metadata-specifications/{emso_version}/OceanSites_codes.md"
emso_codes_url = f"https://raw.githubusercontent.com/emso-eric/emso-metadata-specifications/{emso_version}/EMSO_codes.md"

sdn_vocab_p01 = "https://vocab.nerc.ac.uk/downloads/publish/P01.json"
sdn_vocab_p02 = "https://vocab.nerc.ac.uk/collection/P02/current/?_profile=nvs&_mediatype=application/ld+json"
sdn_vocab_p06 = "https://vocab.nerc.ac.uk/collection/P06/current/?_profile=nvs&_mediatype=application/ld+json"
sdn_vocab_p07 = "https://vocab.nerc.ac.uk/collection/P07/current/?_profile=nvs&_mediatype=application/ld+json"
sdn_vocab_l05 = "https://vocab.nerc.ac.uk/collection/L05/current/?_profile=nvs&_mediatype=application/ld+json"
sdn_vocab_l06 = "https://vocab.nerc.ac.uk/collection/L06/current/?_profile=nvs&_mediatype=application/ld+json"
sdn_vocab_l22 = "https://vocab.nerc.ac.uk/collection/L22/current/?_profile=nvs&_mediatype=application/ld+json"
sdn_vocab_l35 = "https://vocab.nerc.ac.uk/collection/L35/current/?_profile=nvs&_mediatype=application/ld+json"
# standard_names = "https://vocab.nerc.ac.uk/standard_name/?_profile=nvs&_mediatype=application/ld+json"

edmo_codes = "https://edmo.seadatanet.org/sparql/sparql?query=SELECT%20%3Fs%20%3Fp%20%3Fo%20WHERE%20%7B%20%0D%0A%0" \
             "9%3Fs%20%3Fp%20%3Fo%20%0D%0A%7D%20LIMIT%201000000&accept=application%2Fjson"

spdx_licenses_github = "https://raw.githubusercontent.com/spdx/license-list-data/main/licenses.md"

# Copernicus INS TAC Parameter list v3.2
copernicus_param_list = "https://archimer.ifremer.fr/doc/00422/53381/108480.xlsx"


def process_markdown_file(file) -> (dict, dict):
    """
    Processes the Markdown file and parses their tables. Every table is returned as a pandas dataframe.
    :returns: a dict wher keys are table titles and values are dataframes with the info
    """
    with open(file, encoding="utf-8") as f:
        lines = f.readlines()

    title = ""
    tables = {}
    in_table = False
    lines += "\n"  # add an empty line to force table end
    linenum = 0
    for line in lines:
        line = line.strip()
        linenum += 1
        if line.startswith("#"):  # store the title
            title = line.strip().replace("#", "").strip()

        elif not in_table and line.startswith("|"):  # header of the table
            if not line.endswith("|"):
                line += "|"  # fix tables not properly formatted
            table = {}
            headers = line.strip().split("|")
            headers = [h.strip() for h in headers][1:-1]

            for header in headers:
                table[header] = []

            in_table = True

        elif in_table and not line.startswith("|"):  # end of the table
            in_table = False
            tables[title] = pd.DataFrame(table)  # store the metadata as a DataFrame

        elif line.startswith("|---"):  # skip the title and body separator (|----|---|---|)
            continue

        elif line.startswith("|"):  # process the row
            if not line.endswith("|"):
                line += "|"  # fix tables not properly formatted
            fields = [f.strip() for f in line.split("|")[1:-1]]
            for i in range(len(fields)):
                if fields[i] in ["false", "False"]:
                    table[headers[i]].append(False)
                elif fields[i] in ["true", "True"]:
                    table[headers[i]].append(True)
                else:
                    table[headers[i]].append(fields[i])
    return tables


def get_sdn_jsonld_ids(file):
    with open(file, encoding="utf-8") as f:
        data = json.load(f)
    ids = []
    for element in data["@graph"]:
        if "identifier" in element.keys():
            ids.append(element["identifier"])
    return ids


def get_sdn_jsonld_pref_label(file):
    with open(file, encoding="utf-8") as f:
        data = json.load(f)

    names = []
    for element in data["@graph"][1:]:
        if "prefLabel" in element.keys() and "@value" in element["prefLabel"].keys():
            names.append(element["prefLabel"]["@value"])
    return names


def get_sdn_jsonld_uri(file):
    with open(file, encoding="utf-8") as f:
        data = json.load(f)

    names = []
    for element in data["@graph"][1:]:
        if "@id" in element.keys():
            names.append(element["@id"])
    return names


def get_edmo_codes(file):
    with open(file, encoding="utf-8") as f:
        data = json.load(f)

    codes = []
    uris = []
    names = []
    for element in data["results"]["bindings"]:
        if element["p"]["value"] == "http://www.w3.org/ns/org#name":

            code = int(element["s"]["value"].split("/")[-1])
            uris.append(element["s"]["value"])
            codes.append(code)
            names.append(element["o"]["value"])

    df = pd.DataFrame({
        "uri": uris,
        "code": codes,
        "name": names,
    })
    return df

def parse_sdn_jsonld(filename):
    """
    Opens a JSON-LD file from SeaDataNet and try to process it.
    :param filename: file path
    :returns: data (dict), narrower (list), broader (list), related (list)
    """
    with open(filename, encoding="utf-8") as f:
        contents = json.load(f)

    data = {
        "uri": [],
        "identifier": [],
        "prefLabel": [],
        "definition": []
    }

    alias = {  # key-> key to be stored in data dict, value -> all possible keys found in JSON-LD docs
        "definition": ["definition", "skos:definition"],
        "prefLabel": ["definition", "skos:prefLabel"],
        "identifier": ["dc:identifier", "dce:identifier"],
        "uri": ["@id"]
    }

    def get_value_by_alias(mydict, mykey):
        if mykey not in alias.keys():
            return mydict[mykey]
        for try_alias in alias[mykey]:
            try:
                return mydict[try_alias]
            except KeyError:
                pass
        return None

    narrower = {}
    broader = {}
    related = {}
    for element in contents["@graph"]:
        uri = element["@id"]
        if element["@type"] != "skos:Concept":
            continue

        for key in data.keys():
            value = get_value_by_alias(element, key)
            if type(value) == type(None):
                # Check that it is explicitly NoneType
                continue
            if type(value) is dict:
                value = value["@value"]
            data[key].append(value)

        # Initialize as empty list
        narrower[uri] = []
        broader[uri] = []
        related[uri] = []

        def extract_related_elements(mydict, mykeys):
            for mykey in mykeys:
                if mykey not in mydict.keys():
                    continue
                if isinstance(mydict[mykey], dict):
                    return [mydict[mykey]["@id"]]  # generate a dict with the dict value
                elif isinstance(mydict[mykey], list):
                    newlist = []
                    for nested_value in mydict[mykey]:
                        if isinstance(nested_value, dict):
                            newlist.append(nested_value["@id"])
                        else:
                            newlist.append(nested_value)
                    return newlist
                elif isinstance(mydict[mykey], str):
                    return [mydict[mykey]]  # generate a list with the string
                else:
                    raise ValueError(f"Type {type(mydict[mykey])} not expected")
            return []

        # If present, store relationships
        narrower[uri] = extract_related_elements(element, ["skos:narrower", "narrower"])
        broader[uri] = extract_related_elements(element, ["skos:broader", "broader"])
        related[uri] = extract_related_elements(element, ["skos:related", "related"])

    # Remove prefixes like skos and dce
    prefixes = ["skos:", "dce:", "dc:"]
    for p in prefixes:
        for key in list(data.keys()):
            if key.startswith(p):
                new_key = key.replace(p, "")
                data[new_key] = data.pop(key)

    if "@id" in data.keys():
        data["uri"] = data.pop("@id")
    if "identifier" in data.keys():
        data["id"] = data.pop("identifier")

    return data, narrower, broader, related


class EmsoMetadata:
    def __init__(self, force_update=False):

        os.makedirs(".emso", exist_ok=True)  # create a conf dir to store Markdown and other stuff
        os.makedirs(os.path.join(".emso", "jsonld"), exist_ok=True)
        os.makedirs(os.path.join(".emso", "relations"), exist_ok=True)
        ssl._create_default_https_context = ssl._create_unverified_context

        emso_metadata_file = os.path.join(".emso", "EMSO_metadata.md")
        oceansites_file = os.path.join(".emso", "OceanSites_codes.md")
        emso_sites_file = os.path.join(".emso", "EMSO_codes.md")
        sdn_vocab_p01_file = os.path.join(".emso", "jsonld", "sdn_vocab_p01.json")
        sdn_vocab_p02_file = os.path.join(".emso", "jsonld", "sdn_vocab_p02.json")
        sdn_vocab_p06_file = os.path.join(".emso", "jsonld", "sdn_vocab_p06.json")
        sdn_vocab_p07_file = os.path.join(".emso", "jsonld", "sdn_vocab_p07.json")
        sdn_vocab_l05_file = os.path.join(".emso", "jsonld", "sdn_vocab_l05.json")
        sdn_vocab_l06_file = os.path.join(".emso", "jsonld", "sdn_vocab_l06.json")
        sdn_vocab_l22_file = os.path.join(".emso", "jsonld", "sdn_vocab_l22.json")
        sdn_vocab_l35_file = os.path.join(".emso", "jsonld", "sdn_vocab_l35.json")
        edmo_codes_jsonld = os.path.join(".emso", "edmo_codes.json")
        spdx_licenses_file = os.path.join(".emso", "spdx_licenses.md")
        copernicus_params_file = os.path.join(".emso", "copernicus_param_list.xlsx")

        tasks = [
            [emso_metadata_url, emso_metadata_file, "EMSO metadata"],
            [oceansites_codes_url, oceansites_file, "OceanSites"],
            [emso_codes_url, emso_sites_file, "EMSO codes"],
            [sdn_vocab_p01, sdn_vocab_p01_file, "SDN Vocab P01"],
            [sdn_vocab_p02, sdn_vocab_p02_file, "SDN Vocab P02"],
            [sdn_vocab_p06, sdn_vocab_p06_file, "SDN Vocab P06"],
            [sdn_vocab_p07, sdn_vocab_p07_file, "SDN Vocab P07"],
            [sdn_vocab_l05, sdn_vocab_l05_file, "SDN Vocab L05"],
            [sdn_vocab_l06, sdn_vocab_l06_file, "SDN Vocab L06"],
            [sdn_vocab_l22, sdn_vocab_l22_file, "SDN Vocab L22"],
            [sdn_vocab_l35, sdn_vocab_l35_file, "SDN Vocab L35"],
            [edmo_codes, edmo_codes_jsonld, "EDMO codes"],
            [spdx_licenses_github, spdx_licenses_file, "spdx licenses"],
            [copernicus_param_list, copernicus_params_file, "spdx licenses"]
        ]

        download_files(tasks)

        tables = process_markdown_file(emso_metadata_file)
        self.global_attr = tables["Global Attributes"]
        self.variable_attr = tables["Variable Attributes"]
        self.dimension_attr = tables["Dimension Attributes"]
        self.qc_attr = tables["Quality Control Attributes"]
        self.technical_attr = tables["Technical Variables"]

        tables = process_markdown_file(oceansites_file)
        self.oceansites_sensor_mount = list(tables["Sensor Mount"]["sensor_mount"].values)
        self.oceansites_sensor_orientation = list(tables["Sensor Orientation"]["sensor_orientation"].values)
        self.oceansites_data_modes = list(tables["Data Modes"]["Value"].values)
        self.oceansites_data_types = list(tables["Data Types"]["Data type"].values)

        tables = process_markdown_file(emso_sites_file)
        self.emso_regional_facilities = list(tables["EMSO Regional Facilities"]["EMSO Regional Facilities"].values)
        self.emso_sites = list(tables["EMSO Sites"]["EMSO Site"].values)

        tables = process_markdown_file(spdx_licenses_file)
        t = tables["Licenses with Short Idenifiers"]
        # remove extra '[' ']' in license identifiers
        new_ids = [value.replace("[", "").replace("]", "") for value in t["Short Identifier"]]
        self.spdx_license_names = new_ids
        self.spdx_license_uris = {lic: f"https://spdx.org/licenses/{lic}" for lic in self.spdx_license_names}

        sdn_vocabs = {
            "P01": sdn_vocab_p01_file,
            "P02": sdn_vocab_p02_file,
            "P06": sdn_vocab_p06_file,
            "P07": sdn_vocab_p07_file,
            "L05": sdn_vocab_l05_file,
            "L06": sdn_vocab_l06_file,
            "L22": sdn_vocab_l22_file,
            "L35": sdn_vocab_l35_file
        }

        self.sdn_vocabs_ids = {}
        self.sdn_vocabs_pref_label = {}
        self.sdn_vocabs_uris = {}
        self.sdn_vocabs = {}
        self.sdn_vocabs_narrower = {}
        self.sdn_vocabs_broader = {}
        self.sdn_vocabs_related = {}

        t = time.time()
        # Process raw SeaDataNet JSON-ld files and store them sliced in short JSON files
        for vocab, jsonld_file in sdn_vocabs.items():
            csv_filename = os.path.join(".emso", f"{vocab}.csv")
            frelated = os.path.join(".emso", "relations",  f"{vocab}.related")
            fnarrower = os.path.join(".emso", "relations",  f"{vocab}.narrower")
            fbroader = os.path.join (".emso", "relations",  f"{vocab}.broader")
            if os.path.exists(csv_filename):
                df = pd.read_csv(csv_filename)
                with open(frelated) as f:
                    related = json.load(f)
                with open(fnarrower) as f:
                    narrower = json.load(f)
                with open(fbroader) as f:
                    broader = json.load(f)
            else:
                rich.print(f"Loading SDN {vocab}...", end="")
                df, narrower, broader, related = self.load_sdn_vocab(jsonld_file)
                rich.print("[green]done!")
                for filename, values in {fnarrower: narrower, fbroader: broader, frelated: related}.items():
                    with open(filename, "w") as f:
                        json.dump(values, f)
            # for vocab, df in self.sdn_vocabs.items():
                # Storing to CSV to make it easier to search
                df = df[["id", "uri", "prefLabel", "definition"]]
                filename = os.path.join(".emso", f"{vocab}.csv")
                df.to_csv(filename, index=False)

            self.sdn_vocabs[vocab] = df
            self.sdn_vocabs_narrower[vocab] = narrower
            self.sdn_vocabs_broader[vocab] = broader
            self.sdn_vocabs_related[vocab] = related
            self.sdn_vocabs_pref_label[vocab] = df["prefLabel"].values
            self.sdn_vocabs_ids[vocab] = df["id"].values
            self.sdn_vocabs_uris[vocab] = df["uri"].values

        edmo_csv = os.path.join(".emso", f"edmo_codes.csv")
        if not os.path.exists(edmo_csv):
            self.edmo_codes = get_edmo_codes(edmo_codes_jsonld)
            self.edmo_codes.to_csv(edmo_csv, index=False)
        else:
            self.edmo_codes = pd.read_csv(edmo_csv)

        # TODO: Move hardcoded list to OceanSITES_codes.md
        self.oceansites_param_codes = ["AIRT", "CAPH", "CDIR", "CNDC", "CSPD", "DEPTH", "DEWT", "DOX2", "DOXY",
                                       "DOXY_TEMP", "DYNHT", "FLU2", "HCSP", "HEAT", "ISO17", "LW", "OPBS", "PCO2",
                                       "PRES", "PSAL", "RAIN", "RAIT", "RELH", "SDFA", "SRAD", "SW", "TEMP", "UCUR",
                                       "UWND", "VAVH", "VAVT", "VCUR", "VDEN", "VDIR", "VWND", "WDIR", "WSPD"]
        # Convert P02 IDs to 4-letter codes
        self.sdn_p02_names = [code.split(":")[-1] for code in self.sdn_vocabs_ids["P02"]]

        # Parse Copernicus Params excel file
        df = pd.read_excel(copernicus_params_file, sheet_name="Parameters", keep_default_na=False, header=1)
        variables = df["variable name"].dropna().values
        variables = [v.split(" (")[0] for v in variables]  # remove citations
        variables = [v for v in variables if len(v) > 1]   # remove empty lines
        self.copernicus_variables = variables


    @staticmethod
    def clear_downloads():
        """
        Clears all files in .emso folder
        """
        files = get_file_list(".emso")
        for f in files:
            if os.path.isfile(f):
                os.remove(f)

    @staticmethod
    def load_sdn_vocab(filename):
        """
        Loads a SDN vocab into a pandas dataframe.
        """
        data, narrower, broader, related = parse_sdn_jsonld(filename)

        try:
            df = pd.DataFrame(data)
        except Exception as e:
            import rich
            rich.print(data.keys())
            rich.print(f"[orange1]Hey!")
            for key, value in data.items():
                rich.print(f"{key}: {len(value)}")
            raise e
        df = df.rename(columns={"@id": "uri", "dc:identifier": "id"})
        return df, narrower, broader, related

    @staticmethod
    def harmonize_uri(uri):
        """
        Takes a SDN URI and make sure that uses http instead of https and that it finishes with a /
        """
        if uri.startswith("https"):
            uri = uri.replace("https", "http")

        if not uri.endswith("/"):
            uri += "/"
        return uri

    def vocab_get(self, vocab_id, uri, key):
        """
        Search in vocab <vocab_id> for the element with matching uri and return element identified by key
        """
        uri = self.harmonize_uri(uri)
        __allowed_keys = ["prefLabel", "id", "definition"]
        if key not in __allowed_keys:
            raise ValueError(f"Key '{key}' not valid, allowed keys: {__allowed_keys}")

        df = self.sdn_vocabs[vocab_id]
        row = df.loc[df["uri"] == uri]
        if row.empty:
            #raise LookupError(f"Could not get {key} for '{uri}' in vocab {vocab_id}")
            rich.print(f"[red]Could not get {key} for '{uri}' in vocab {vocab_id}")
            return ""
        return row[key].values[0]

    def vocab_get_by_urn(self, vocab_id, urn, key):
        """
        Search in vocab <vocab_id> for the element with matching uri and return element identified by key
        """
        __allowed_keys = ["prefLabel", "uri", "definition"]
        if key not in __allowed_keys:
            raise ValueError(f"Key '{key}' not valid, allowed keys: {__allowed_keys}")

        df = self.sdn_vocabs[vocab_id]
        row = df.loc[df["id"] == urn]
        if row.empty:
            raise LookupError(f"Could not get {key} for '{urn}' in vocab {vocab_id}")
        return row[key].values[0]

    def get_relations(self, vocab_id, uri, relation, target_vocab):
        """
        Takes a relation list from a vocabulary (narrower, broader or related), looks for a term identified by URI and
        returns a list of all the terms within that relationtship that are from the vocabuary target_vocab.

        This function is useful to get related metadata from a term, from instance

            "P01", <param>, "related", "P06" -> get the prefered units for a parameter listed within P01
            "L22", <model>, "related", "L35" -> get the  manufacturer of a sensor
            "P02", <param>, "narrower", "P01" -> get all possible fine-grained parameters values from a braod parameter

        :param vocab_id: ID of the vocabulary being used
        :param uri: URI of the term whose relations will be explored
        :param relation: type of relationship, possible values are narrower, broader and related
        :param target_vocab: id of the vocabulary terms that we want to find
        :returns: list with matches
        """
        __valid_relations = ["narrower", "broader", "related"]
        uri = self.harmonize_uri(uri)

        if relation not in __valid_relations:
            raise LookupError(f"Not a valid relation: '{relation}', expected one of '{__valid_relations} ")

        if relation == "narrower":
            relations = self.sdn_vocabs_narrower[vocab_id]
        elif relation == "broader":
            relations = self.sdn_vocabs_broader[vocab_id]
        else:  # related
            relations = self.sdn_vocabs_related[vocab_id]

        try:
            uri_relations = relations[uri]
        except KeyError:
            rich.print(f"[red]relation {relation} for {uri} not found!")
            return ""

        if type(uri_relations) is str:  # make sure it's a list
            uri_relations = [uri_relations]

        results = []
        for term in uri_relations:
            if target_vocab in term:
                results.append(term)
        return results

    def get_relation(self, vocab_id, uri, relation, target_vocab):
        """
        The same as get relations but throws an error if more than one element are found
        """
        results = self.get_relations(vocab_id, uri, relation, target_vocab)
        if len(results) == 0:
            rich.print(f"[red]Could not find relation {relation} for {uri}")
            return ""
        elif len(results) != 1:
            raise LookupError(f"Expected 1 value, got {len(results)}")

        return results[0]



#!/usr/bin/env python3
"""
Re-implementation of the WaterFrame class. Originally implemented in the mooda package, it is no longer maintained, so
a custom re-implementation is included here.

author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 6/6/24
"""
import pandas as pd


class WaterFrame:
    def __init__(self, data: pd.DataFrame, metadata: dict, vocabulary: dict):
        """
        This class is a lightweight re-implementation of WaterFrames, originally from mooda package. It has been
        reimplemented due to lack of maintenance of the original package.
        """
        assert type(data) is pd.DataFrame
        assert type(metadata) is dict
        assert type(vocabulary) is dict

        self.data = data  # Here, we should have a dataframe
        self.metadata = metadata
        self.vocabulary = vocabulary

        # Now make sure that all variables have an entry in the vocabulary

        for col in self.data.columns:
            assert col in self.vocabulary.keys(), f"Vocabulary dict does not have netry for {col}"



#!/usr/bin/env python3
"""

author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 29/4/23
"""
import logging

from .waterframe import WaterFrame
import pandas as pd
from .constants import dimensions, qc_flags, fill_value
import numpy as np
import rich
import netCDF4 as nc

from .metadata_templates import dimension_metadata, quality_control_metadata
from .netcdf import wf_to_multidim_nc, read_nc
from .utils import drop_duplicates, merge_dicts


def get_variables(wf):
    """
    returns a list of QC variables within a waterframe
    """
    vars = []
    dimensions_l = [d.lower() for d in dimensions]
    for c in wf.data.columns:
        if not c.endswith("_QC") and not c.endswith("_STD") and c.lower() not in dimensions and c.lower() not in dimensions_l:
            vars.append(c)
    return vars


def get_dimensions(wf):
    """
    returns a list of QC variables within a waterframe
    """
    return [col for col in wf.data.columns if col.upper() in dimensions]


def get_qc_variables(wf):
    """
    returns a list of QC variables within a waterframe
    """
    return [col for col in wf.data.columns if col.endswith("_QC")]


def get_std_variables(wf):
    """
    returns a list of standard deviation variables within a waterframe
    """
    return [col for col in wf.data.columns if col.endswith("_STD")]


def harmonize_dataframe(df, fill_value=fill_value):
    """
    Takes a dataframe and harmonizes all variable names. All vars are converter to upper case except for lat, lon
    and depth.All QC and STD vars are put to uppercase.
    """
    # harmonize time
    for time_key in ["time", "timestamp", "datetime", "date time"]:
        for key in df.columns:
            if key.lower() == time_key:
                df = df.rename(columns={key: "TIME"})

    for var in df.columns:
        skip = False
        for dim in dimensions:  # skip all dimensions and QC related to dimensions
            if var.startswith(dim):
                skip = True
        if not skip:
            df = df.rename(columns={var: var.upper()})

    # make sure that _QC are uppercase
    for var in df.columns:
        if var.lower().endswith("_qc"):
            df = df.rename(columns={var: var[:-3] + "_QC"})

    # make sure that _QC are uppercase
    for var in df.columns:
        if var.lower().endswith("_std"):
            df = df.rename(columns={var: var[:-4] + "_STD"})

    missing_data = qc_flags["missing_value"]
    for col in df.columns:
        # make sure no NaNs are present in the dataframe
        if col.endswith("_QC"):
            if df[col].dtype != int:
                df[col] = df[col].replace(np.nan, missing_data)  # instead of nan put missing value
                df[col] = df[col].astype(int)

        # replace NaN by fill value
        else:
            if df[col].isnull().any():
                df[col] = df[col].replace(np.nan, fill_value)

    return df


# -------- Functions to handle data from CSV files -------- #
def load_csv_data(filename, sep=",") -> (pd.DataFrame, list):
    """
    Loads data from a CSV file and returns a WaterFrame
    """
    if not filename.endswith(".csv"):
        rich.print(f"[yellow]WARNING! extension of file {filename} is not '.csv', trying anyway...")

    header_lines = csv_detect_header(filename, separator=sep)
    df = pd.read_csv(filename, skiprows=header_lines, sep=sep)
    df = df_force_upper_case(df)
    wf = df_to_wf(df)
    wf.metadata["$datafile"] = filename  # Add the filename as a special param
    return wf


def df_to_wf(df: pd.DataFrame) -> WaterFrame:
    """
    Converts a dataframe into a waterframe
    """
    df = harmonize_dataframe(df)
    vocabulary = {c: {} for c in df.columns}
    wf = WaterFrame(df, {}, vocabulary)
    wf.data["TIME"] = pd.to_datetime(wf.data["TIME"])
    return wf




def csv_detect_header(filename, separator=","):
    """
    Opens a CSV, reads the last 3 lines and extracts the number of fields. Then it goes back to the beginning and
    detects the first line which is not a header
    """
    with open(filename) as f:
        lines = f.readlines()

    if len(lines) < 3:
        # empty CSV, first line is the header
        return 0

    nfields = len(lines[-2].split(separator))
    if nfields < 2 or not (nfields == len(lines[-3].split(separator)) == len(lines[-4].split(separator))):
        raise ValueError("Could not determine number of fields")

    # loop until a first a line with nfields is found
    i = 0
    while len(lines[i].split(separator)) != nfields:
        i += 1
    return i


def wf_force_upper_case(wf: WaterFrame) -> WaterFrame:
    # Force upper case in dimensions
    for key in wf.data.columns:
        if key.upper() in dimensions and key.upper() != key:
            wf.data = wf.data.rename(columns={key: key.upper()})
            wf.vocabulary[key.upper()] = wf.vocabulary.pop(key)
    return wf


def df_force_upper_case(df: pd.DataFrame) -> pd.DataFrame:
    # Force upper case in dimensions
    for key in df.columns:
        if key.upper() in dimensions and key.upper() != key:
            df = df.rename(columns={key: key.upper()})
    return df


def load_data(file: str):
    """
    Opens a CSV or NetCDF data and returns a WaterFrame
    """
    if file.endswith(".csv"):
        wf = load_csv_data(file)
    elif file.endswith(".nc"):
        wf = load_nc_data(file)
    else:
        raise ValueError("Unimplemented file format for data!")

    return wf


def semicolon_to_list(attr: str):
    """
    Converts semi-colon separated list of items into a python list
    """
    if type(attr) == str and ";" in attr:
        return attr.split(";")
    else:
        return attr


# -------- Load NetCDF data -------- #
def load_nc_data(filename, drop_duplicates=False, process_lists=True) -> (WaterFrame, list):
    """
    Loads NetCDF data into a waterframe
    """
    wf = read_nc(filename, decode_times=False)
    if process_lists:  # Process semicolon separated lists
        for key, value in wf.metadata.items():
            wf.metadata[key] = semicolon_to_list(value)

        for var in wf.vocabulary.keys():
            for key, value in wf.vocabulary[var].items():
                wf.vocabulary[var][key] = semicolon_to_list(value)
    wf.data = wf.data.reset_index()

    if "row" in wf.data.columns:
        # a 'row' column may be introduced by reset index if there previous index was an integer
        del wf.data["row"]
    wf = wf_force_upper_case(wf)
    df = wf.data
    units = wf.vocabulary["TIME"]["units"]
    if "since" not in units:  # netcdf library requires that the units fields has the 'since' keyword
        if "sdn_parameter_urn" in wf.vocabulary["TIME"].keys() and wf.vocabulary["TIME"]["sdn_parameter_urn"] == "SDN:P01::ELTJLD01":
            units = "days since 1950-01-01T00:00:00z"
        else:
            units = "seconds since 1970-01-01T00:00:00z"
    df["TIME"] = nc.num2date(df["TIME"].values, units, only_use_python_datetimes=True, only_use_cftime_datetimes=False)
    df["TIME"] = pd.to_datetime((df["TIME"]), utc=True)
    if drop_duplicates:
        dups = df[df["TIME"].duplicated()]
        if len(dups) > 0:
            rich.print(f"[yellow]WARNING! detected {len(dups)} duplicated times!, deleting")
            df = drop_duplicates(df)

    wf.data = df  # assign data
    wf.metadata["$datafile"] = filename  # Add the filename as a special param

    # make sure that every column in the dataframe has an associated vocabulary
    for varname in wf.data.columns:
        if varname not in wf.vocabulary.keys():
            rich.print(f"[red]ERROR: Variable {varname} not listed in metadata!")
            wf.vocabulary[varname] = {}  # generate empty metadata vocab
    return wf


# -------- Coordinate-related functions -------- #
def add_coordinates(wf: WaterFrame, latitude, longitude, depth):
    """
    Takes a waterframe and adds nominal lat/lon/depth values
    """
    coordinates = {"LATITUDE": latitude, "LONGITUDE": longitude, "DEPTH": depth}
    for name, value in coordinates.items():
        if name not in wf.data.columns:
            wf.data[name] = value
            wf.data[f"{name}_QC"] = qc_flags["nominal_value"]
            wf.vocabulary[name] = dimension_metadata(name)
            wf.vocabulary[f"{name}_QC"] = quality_control_metadata(wf.vocabulary[name]["long_name"])
    return wf


def ensure_coordinates(wf, required=["DEPTH", "LATITUDE", "LONGITUDE"]):
    """
    Make sure that depth, lat and lon variables (and their QC) are properly set
    """
    error = False
    df = wf.data
    for r in required:
        if r not in df.columns:
            error = True
            rich.print(f"[red]Coordinate {r} is missing!")
        if df[r].dtype != np.float64:
            df[r] = df[r].astype(np.float64)

    if error:
        raise ValueError("Coordinates not properly set")


def update_waterframe_metadata(wf: WaterFrame, meta: dict):
    """
    Merges a full metadata JSON dict into a Waterframe
    """
    wf.metadata = merge_dicts(meta["global"], wf.metadata)
    wf.vocabulary = merge_dicts(meta["variables"], wf.vocabulary)

    keywords = get_variables(wf)
    wf.metadata["keywords"] = keywords
    wf.metadata["keywords_vocabulary"] = "SeaDataNet Parameter Discovery Vocabulary"

    # Updating ancillary variables with QC and STD data
    for qc in get_qc_variables(wf):
        varname = qc.replace("_QC", "")
        varmeta = wf.vocabulary[varname]
        if "ancillary_variables" not in varmeta.keys():
            varmeta["ancillary_variables"] = []
        varmeta["ancillary_variables"].append(qc)

    for std in get_std_variables(wf):
        varname = std.replace("_STD", "")
        varmeta = wf.vocabulary[varname]
        if "ancillary_variables" not in varmeta.keys():
            varmeta["ancillary_variables"] = []
        varmeta["ancillary_variables"].append(std)

    # Update variable coordinates with the dataframe dimensions
    for var in get_variables(wf):
        wf.vocabulary[var]["coordinates"] = dimensions

    # check if all fields are filled, otherwise set a blank string
    __global_attr = ["doi", "platform_code", "wmo_platform_code"]
    for attr in __global_attr:
        if attr not in wf.metadata.keys():
            wf.metadata[attr] = ""

    __variable_fields = ["reference_scale", "comment"]
    for attr in __variable_fields:
        for varname in get_variables(wf):
            if attr not in wf.vocabulary[varname].keys():
                wf.vocabulary[varname][attr] = ""

    return wf


def all_equal(values: list):
    """
    checks if all elements in a list are equal
    :param values: input list
    :returns: True/False
    """
    baseline = values[0]
    equals = True
    for element in values[1:]:
        if element != baseline:
            equals = False
            break
    return equals


def consolidate_metadata(dicts: list) -> dict:
    """
    Consolidates metadata in a list of dicts. All dicts are expected to have the same fields. If all the values are
    equal, keep a single value. If the values are not equal, create a list. However, if it is a sensor_* key, all
    values will be kept.
    """
    keys = [key for key in dicts[0].keys()]  # Get the keys from the first dictionary
    final = {}
    for key in keys:
        values = [d[key] for d in dicts]  # get all the values
        if all_equal(values) and not key.startswith("sensor_"):
            final[key] = values[0]  # get the first element only, all are the same!
        else:
            final[key] = values  # put the full list
    return final


def set_multisensor(wf: WaterFrame):
    """
    Looks through all the variables and checks if data comes from two or more sensors. Sets the multisensor flag
    """
    serial_numbers = []
    for varname, varmeta in wf.vocabulary.items():
        if "sensor_serial_number" not in varmeta.keys():
            continue  # avoid QC and STD vars

        if type(varmeta["sensor_serial_number"]) == str:
            if varmeta["sensor_serial_number"] not in serial_numbers:
                serial_numbers.append(varmeta["sensor_serial_number"])
        elif type(varmeta["sensor_serial_number"]) == list:
            for serial in varmeta["sensor_serial_number"]:
                if serial not in serial_numbers:
                    serial_numbers.append(serial)
    if len(serial_numbers) > 1:
        multi_sensor = True
    elif len(serial_numbers) == 1:
        multi_sensor = False
    else:
        wf.metadata["$multisensor"] = False
        raise LookupError("No sensor serial numbers found!!")
    wf.metadata["$multisensor"] = multi_sensor

    return wf


def export_to_netcdf(wf, filename):
    """
    Stores the waterframe to a NetCDF file
    """
    # If only one sensor remove all sensor_id fields
    set_multisensor(wf)
    if not wf.metadata['$multisensor']:
        if "SENSOR_ID" in wf.data.columns:
            del wf.data["SENSOR_ID"]
        if "SENSOR_ID" in wf.vocabulary.keys():
            del wf.vocabulary["SENSOR_ID"]
        dimensions.remove("SENSOR_ID")

    # Remove internal elements in metadata
    [wf.metadata.pop(key) for key in wf.metadata.copy().keys() if key.startswith("$")]
    wf_to_multidim_nc(wf, filename, dimensions, fill_value=fill_value, time_key="TIME", join_attr=";")


def extract_netcdf_metadata(wf):
    """
    Extracts data from a waterframe into .full.json data
    """
    metadata = {
        "global": wf.metadata,
        "variables": wf.vocabulary
    }
    for key in list(metadata["global"].keys()):
        if key.startswith("$"):
            del metadata["global"][key]  # remove special fields

    return metadata


def get_netcdf_metadata(filename):
    """
    Returns the metadata from a NetCDF file
    :param: filename
    :returns: dict with the metadata { "global": ..., "variables": {"VAR1": {...},"VAR2":{...}}
    """
    wf = load_nc_data(filename, process_lists=False)
    metadata = {
        "global": wf.metadata,
        "variables": wf.vocabulary
    }
    return metadata


#!/usr/bin/env python3
"""
Functions to autofill missing metadata

author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 29/4/23
"""

from argparse import ArgumentParser
import json
import rich
from . import EmsoMetadata
from .constants import dimensions, iso_time_format
from .dataset import get_variables, set_multisensor
from .metadata_templates import choose_interactively, dimension_metadata, quality_control_metadata
from .utils import merge_dicts
from .waterframe import WaterFrame


def set_default(m: dict, key: str, value: any):
    """
    Sets the key:value if key doesn't exist or current value is null
    """
    if key not in m.keys():
        m[key] = value
    elif not m[key]:
        m[key] = value
    return m


def autofill_minmeta(minmeta: dict, emso: EmsoMetadata):
    """
    Takes a minimal metadata JSON dict and tries to fill the gaps with default values
    """
    set_default(minmeta["global"], "~Conventions", ["OceanSITES", "EMSO"])
    set_default(minmeta["global"], "~format_version", "1.4")
    set_default(minmeta["global"], "~update_interval", "void")
    set_default(minmeta["global"], "~network", "EMSO")
    set_default(minmeta["global"], "~license", "CC-BY-4.0")
    for varname, varmeta in minmeta["variables"].items():
        sdn_parameter_uri = emso.harmonize_uri(varmeta["*sdn_parameter_uri"])
        sdn_uom_uri = emso.get_relation("P01", sdn_parameter_uri, "related", "P06")
        set_default(varmeta, "~sdn_uom_uri", sdn_uom_uri)

        if "~standard_name" not in varmeta.keys() or not varmeta["~standard_name"]:
            standard_name_uris = emso.get_relations("P01", sdn_parameter_uri, "broader", "P07")
            if len(standard_name_uris) == 1:
                standard_name_uri = standard_name_uris[0]  # Match!
                standard_name = emso.vocab_get("P07", standard_name_uri, "prefLabel")
            elif len(standard_name_uris) > 1:
                standard_names = [emso.vocab_get("P07", s, "prefLabel") for s in standard_name_uris]
                # Let the user choose interactively which one
                standard_name = choose_interactively("standard_name", "", standard_names)
            else:
                # No P07 links found for this URI, so instead of P01 try to use the broader P02
                p02_uri = emso.get_relation("P01", sdn_parameter_uri, "broader", "P02")
                standard_name_uris = emso.get_relations("P02", p02_uri, "narrower", "P07")
                if len(standard_name_uris) == 1:
                    standard_name_uri = standard_name_uris[0]  # Match!
                    standard_name = emso.vocab_get("P07", standard_name_uri, "prefLabel")
                elif len(standard_name_uris) > 1:
                    standard_names = [emso.vocab_get("P07", s, "prefLabel") for s in standard_name_uris]
                    # Let the user choose interactively which one
                    standard_name = choose_interactively("standard_name", varname, standard_names)
                else:
                    rich.print("[red]Could not deduce any standard_name!")
                    standard_name = ""
            set_default(varmeta, "~standard_name", standard_name)

    return minmeta


def expand_minmeta(wf: WaterFrame, minmeta: dict, emso: EmsoMetadata) -> dict:
    """
    Expands minimal metadata into full metadata and sotres it within the WaterFrame
    """
    metadata = minmeta.copy()

    if "coordinates" in metadata.keys():
        del metadata["coordinates"]

    metadata["global"] = autofill_global(metadata["global"], emso)
    metadata["sensor"] = autofill_sensor(metadata["sensor"], emso)

    # Propagate from global sensor metadata to every variable
    metadata, sensor_id = propagate_sensor_metadata(metadata)
    wf.metadata["$sensor_id"] = sensor_id

    # Now, add dimensions info
    for dimname in dimensions:
        metadata["variables"][dimname] = dimension_metadata(dimname)

    # Autofill all variables
    [autofill_variable(v, emso) for name, v in metadata["variables"].items() if name != "SENSOR_ID"]

    # Add QC metadata for all variables and dimensions
    for varname, varmeta in metadata["variables"].copy().items():
        if varname == "SENSOR_ID":
            continue
        qcmeta = quality_control_metadata(varmeta["long_name"])
        metadata["variables"][varname + "_QC"] = qcmeta

    if "$minmeta" in wf.metadata.keys():
        full_meta_file = wf.metadata["$minmeta"].replace(".min.json", ".full.json")
        with open(full_meta_file, "w") as f:
            f.write(json.dumps(metadata, indent=2))
    return metadata


def autofill_variable(varmeta: dict, emso: EmsoMetadata) -> dict:
    """
    Expands the variable metadata adding uris, uoms and names for variables and units
    """

    if "sdn_parameter_uri" in varmeta.keys():
        sdn_parameter_uri = varmeta["sdn_parameter_uri"]
    elif "sdn_parameter_urn" in varmeta.keys():  # find the URI based on the URN
        urn = varmeta["sdn_parameter_urn"]
        sdn_parameter_uri = emso.vocab_get_by_urn("P01", urn, "uri")
    else:
        raise LookupError("URN nor URI present in variable!")

    sdn_parameter_uri = emso.harmonize_uri(sdn_parameter_uri)

    if "sdn_parameter_uri" not in varmeta.keys():
        varmeta["sdn_parameter_uri"] = sdn_parameter_uri 

    label = emso.vocab_get("P01", sdn_parameter_uri, "prefLabel")
    sdn_id = emso.vocab_get("P01", sdn_parameter_uri, "id")
    varmeta["sdn_parameter_urn"] = sdn_id
    varmeta["sdn_parameter_name"] = label.strip()

    if "sdn_uom_uri" in varmeta.keys():
        sdn_uom_uri = varmeta["sdn_uom_uri"]
    else:
        rich.print(f"[yellow]WARNING: units for {sdn_parameter_uri} not set, using P01 default units...")
        sdn_uom_uri = emso.get_relation("P01", sdn_parameter_uri, "related", "P06")

    label = emso.vocab_get("P06", sdn_uom_uri, "prefLabel")
    iden = emso.vocab_get("P06", sdn_uom_uri, "id")
    varmeta["sdn_uom_uri"] = sdn_uom_uri
    varmeta["sdn_uom_urn"] = iden
    varmeta["sdn_uom_name"] = label.strip()
    varmeta["units"] = label
    return varmeta


def autofill_global(m: dict, emso: EmsoMetadata) -> dict:
    if "Conventions" not in m.keys():
        m["Conventions"] = ["EMSO ERIC", "OceanSITES"]

    # Getting EDMO URL from the code
    edmo_code = m["institution_edmo_code"]
    edmo_uri = f"https://edmo.seadatanet.org/report/{edmo_code}"
    m = set_default(m, "institution_edmo_code", edmo_code)
    df = emso.edmo_codes
    institution_name = df.loc[df["uri"] == edmo_uri]["name"].values[0]
    m = set_default(m, "institution_edmo_uri", f"https://edmo.seadatanet.org/report/{edmo_code}")
    m = set_default(m, "institution", institution_name)
    m["license_uri"] = emso.spdx_license_uris[m["license"]]
    return m


def autofill_sensor(s: dict, emso: EmsoMetadata) -> dict:

    if "sensor_model_uri" in s.keys():
        sensor_uri = s["sensor_model_uri"]
    elif "sensor_reference" in s.keys():
        sensor_uri = s["sensor_reference"]
    else:
        raise LookupError("Could not find sensor reference!")

    try:
        s["sensor_model"] = emso.vocab_get("L22", sensor_uri, "prefLabel")
        s["sensor_SeaVoX_L22_code"] = emso.vocab_get("L22", sensor_uri, "id")
    except LookupError:
        rich.print("[red]model not found in L22!")
        s["sensor_SeaVoX_L22_code"] = ""
        s["sensor_manufacturer_uri"] = ""
        s["sensor_manufacturer_urn"] = ""
        s["sensor_manufacturer"] = ""
        return s

    try:
        manufacturer_uri = emso.get_relation("L22", sensor_uri, "related", "L35")
        s["sensor_manufacturer_uri"] = manufacturer_uri
        s["sensor_manufacturer_urn"] = emso.vocab_get("L35", manufacturer_uri, "id")
        s["sensor_manufacturer"] = emso.vocab_get("L35", manufacturer_uri, "prefLabel")
    except LookupError:
        rich.print("[red]No manufacturer found on SeaDataNet L35 vocab!!")
        s["sensor_manufacturer_uri"] = ""
        s["sensor_manufacturer_urn"] = ""
        s["sensor_manufacturer"] = ""

    if "sensor_model_uri" in s.keys():
        s["sensor_reference"] = s.pop("sensor_model_uri")
    return s


def autofill_waterframe_coverage(wf: WaterFrame) -> WaterFrame:
    """
    Autofills geospatial and time coverage in a WaterFrame
    """
    wf.metadata["geospatial_lat_min"] = wf.data["LATITUDE"].min()
    wf.metadata["geospatial_lat_max"] = wf.data["LATITUDE"].max()
    wf.metadata["geospatial_lon_min"] = wf.data["LONGITUDE"].min()
    wf.metadata["geospatial_lon_max"] = wf.data["LONGITUDE"].min()
    wf.metadata["geospatial_vertical_min"] = int(wf.data["DEPTH"].min())
    wf.metadata["geospatial_vertical_max"] = int(wf.data["DEPTH"].min())
    wf.metadata["time_coverage_start"] = wf.data["TIME"].min().strftime(iso_time_format)
    wf.metadata["time_coverage_end"] = wf.data["TIME"].max().strftime(iso_time_format)
    return wf


def autofill_coordinates(wf: WaterFrame) -> WaterFrame:
    """
    Autofills the coordinates section of each variable
    """
    vars = get_variables(wf)
    for varname in vars:
        varmeta = wf.vocabulary[varname]
        varmeta["coordinates"] = dimensions
    return wf


def propagate_sensor_metadata(metadata: dict) -> (dict, str):
    """
    Takes a metadata dict and propagates the file-wide sensor information to all the variables
    """
    if not "sensor" in metadata.keys():
        return metadata  # no file-wide sensor info

    sensor_meta = metadata["sensor"]
    for varname, varmeta in metadata["variables"].items():
        # Propagate only in variables, not dimensions, QCs or STDs
        if varname not in dimensions and not varname.endswith("_QC") and not varname.endswith("_STD"):
            varmeta = merge_dicts(sensor_meta, varmeta)
            metadata["variables"][varname] = varmeta

    sensor_id = metadata["sensor"]["sensor_serial_number"]
    del metadata["sensor"]
    return metadata, sensor_id


def autofill_waterframe(wf):
    """
    Takes a waterframe and tries to autofill it
    """
    emso = EmsoMetadata()
    variables = get_variables(wf)

    wf = autofill_coordinates(wf)  # fill the coordinates

    rich.print("Autofilling variables")
    for varname in variables:
        varmeta = wf.vocabulary[varname]
        try:
            wf.vocabulary[varname] = autofill_variable(varmeta, emso)
        except LookupError as e:
            rich.print(f"[red]couldn't autofill {varname} metadata: {e}")

        try:
            wf.vocabulary[varname] = autofill_sensor(varmeta, emso)
        except LookupError as e:
            rich.print(f"[red]couldn't autofill sensor metadata for {varname}: {e}")


    try:
        wf = set_multisensor(wf)
    except LookupError as e:
        rich.print(f"[red]ERROR {e}")

    return wf




#!/usr/bin/env python3
"""
This scripts contains some utilities to work with ERDDAP.

author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 1/3/23
"""

import requests
import rich
import json



class ERDDAP:
    """
    Class that implements common functionalities to perform in ERDDAP servers
    """ 
    def __init__(self, url):
        rich.print(f"User url: {url}")
        # Getting rid of everything after the domain name
        parts = url.split("/")[:3]
        url = "/".join(parts)
        if not url.endswith("/erddap"):
            url += "/erddap"
        rich.print(f"Using url: {url}")
        self.url = url

    @staticmethod
    def get(url,  headers={"Content-Type": "application/json"}):
        r = requests.get(url, headers=headers)
        if r.status_code != 200:
            rich.print(f"[red]HTTP Error: {r.status_code}")
            rich.print(f"[red]{r.text}")
            raise ValueError("exit")
        else:
            return json.loads(r.text)

    def dataset_list(self) -> list:
        """
        Returns a list of dataset IDs
        :return: dataset list
        """
        datasets_url = self.url + "/info/index.json"
        r = self.get(datasets_url)
        datasets = {}
        columns = r["table"]["columnNames"]
        dataset_ids = []
        for row in r["table"]["rows"]:
            dataset = {}
            for i in range(len(row)):
                dataset[columns[i]] = row[i]
            dataset_id = dataset["Dataset ID"]
            if dataset_id != "allDatasets":
                datasets[dataset_id] = dataset
                dataset_ids.append(dataset["Dataset ID"])

        return dataset_ids

    def dataset_metadata(self, dataset_id):
        """
        Formats from ERDDAP's ugly and disgusting JSON format to a nice, well-structured JSON
        :param url: ERDDAP url
        :param dataset_id: ID of the dataset
        :return: a dict with the metadata well structure, like:
            {
            "global": {key:  value, ...},
            "variables": {
                var1: {key, value, ...},
                var2: {key, value, ...},
                ...
                }
            }
        """
        metadata_url = f"{self.url}/info/{dataset_id}/index.json"
        r = self.get(metadata_url)
        metadata = {
            "global": {"dataset_id": dataset_id},
            "variables": {},
            "qc": {}
        }
        for row in r["table"]["rows"]:
            # Each row is similar to: ['attribute', 'NC_GLOBAL', 'author', 'String', 'Enoc Martinez']
            # so 0 -> 'attribute' or 'variable', 1 -> parameter name, 2 -> key, 3 -> data type and 4-> value
            row_type = row[0]
            param = row[1]
            attribute_key = row[2]
            data_type = row[3]
            attribute_value = row[4]

            if row_type == "variable":
                metadata["variables"][param] = {}  # create new dict
            elif row_type == "attribute":
                if param == "NC_GLOBAL":
                    metadata["global"][attribute_key] = attribute_value  # process global attribute
                else:
                    metadata["variables"][param][attribute_key] = attribute_value  # process variable attribute
            else:
                rich.print(f"WARNING could not process row {row}")

        return metadata



#!/usr/bin/env python3
"""

author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 6/6/24
"""
import os
import shutil
import urllib
from argparse import ArgumentParser
import rich
import unittest
import subprocess
import sys
import time
import json
import pandas as pd
import inspect

try:
    from src.emso_metadata_harmonizer import generate_dataset, erddap_config
    from src.emso_metadata_harmonizer.metadata.dataset import load_data
except ModuleNotFoundError:
    # Get the directory of the current script
    current_dir = os.path.dirname(os.path.abspath(__file__))
    # Get the parent directory (project root)
    parent_dir = os.path.abspath(os.path.join(current_dir, os.pardir))
    # Add the parent directory to the sys.path
    sys.path.insert(0, parent_dir)
    from src.emso_metadata_harmonizer import generate_dataset, erddap_config
    from src.emso_metadata_harmonizer.metadata.dataset import load_data
    from src.emso_metadata_harmonizer.metadata.emso import EmsoMetadata


def run_subprocess(cmd):
    """
    Runs a command as a subprocess. If the process retunrs 0 returns True. Otherwise prints stderr and stdout and returns False
    :param cmd: command (list or string)
    :return: True/False
    """
    assert (type(cmd) is list or type(cmd) is str)
    if type(cmd) == list:
        cmd_list = cmd
    else:
        cmd_list = cmd.split(" ")
    proc = subprocess.run(cmd_list, capture_output=True)
    if proc.returncode != 0:
        rich.print(f"\n[red]ERROR while running command '{cmd}'")
        if proc.stdout:
            rich.print(f"subprocess stdout:")
            rich.print(f">[bright_black]    {proc.stdout.decode()}")
        if proc.stderr:
            rich.print(f"subprocess stderr:")
            rich.print(f">[bright_black] {proc.stderr.decode()}")

        raise ValueError(f"subprocess failed: {cmd_list}")


class MetadataHarmonizerTester(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        os.makedirs("erddapData", exist_ok=True)
        rich.print("Starting erddap docker container...")
        run_subprocess("docker compose up -d")

        cls.two_ctds_dataset = "erddapData/2CTDs/2ctds_dataset.nc"
        cls.datasets_default_xml = os.path.join("conf", "datasets_default.xml")
        cls.datasets_xml = os.path.join("conf", "datasets.xml")
        shutil.copy2(cls.datasets_default_xml, cls.datasets_xml)

        cls.dataset_id = "2CTDs"

        cls.erddap_url = "http://localhost:8080/erddap"
        erddap_up = False
        while not erddap_up:
            try:
                urllib.request.urlretrieve(cls.erddap_url)
                erddap_up = True
            except urllib.error.URLError:
                rich.print("waiting for ERDDAP to start...")
                time.sleep(2)
            except ConnectionError:
                rich.print("waiting for ERDDAP to start...")
                time.sleep(2)

    def test_01_create_dataset(self):
        """Creates a dataset based on examples files"""
        rich.print(f"[purple]Running test {inspect.currentframe().f_code.co_name}")
        metadata_files = [
            "../examples/2CTDs/SBE16.min.json",
            "../examples/2CTDs/SBE37.min.json"
        ]

        data_files = [
            "../examples/2CTDs/SBE16.csv",
            "../examples/2CTDs/SBE37.csv"
        ]

        os.makedirs(os.path.dirname(self.two_ctds_dataset), exist_ok=True)
        generate_dataset(data_files, metadata_files, autofill=False, output=self.two_ctds_dataset)
        wf = load_data(self.two_ctds_dataset)
        self.assertTrue(os.path.exists(self.two_ctds_dataset))

    def test_02_create_dataset_from_dict(self):
        """Creates a dataset based on examples files"""
        rich.print(f"[purple]Running test {inspect.currentframe().f_code.co_name}")
        metadata_files = [
            "../examples/2CTDs/SBE16.min.json",
            "../examples/2CTDs/SBE37.min.json"
        ]

        data_files = [
            "../examples/2CTDs/SBE16.csv",
            "../examples/2CTDs/SBE37.csv"
        ]

        metadata = []  # convert metadata files to dicts
        for m in metadata_files:
            with open(m) as f:
                metadata.append(json.load(f))

        # convert data files to DataFrames
        data = [pd.read_csv(f) for f in data_files]

        os.makedirs(os.path.dirname(self.two_ctds_dataset), exist_ok=True)
        generate_dataset(data, metadata, autofill=False, output=self.two_ctds_dataset)
        # Create it two times, to ensure that the original data is not broken!
        generate_dataset(data, metadata, autofill=False, output=self.two_ctds_dataset)

        wf = load_data(self.two_ctds_dataset)
        self.assertTrue(os.path.exists(self.two_ctds_dataset))

        # Now create a dataset with an existing instance of EmsoMetadata
        emso = EmsoMetadata()
        generate_dataset(data, metadata, autofill=False, output=self.two_ctds_dataset, emso_metadata=emso)

        # Now do the same, but delete a mandatory element from metadata, we should get a ValueError
        metadata[0]["global"]["$site_code"] = ""
        with self.assertRaises(ValueError):
            generate_dataset(data, metadata, autofill=False, output=self.two_ctds_dataset)


    def test_03_config_erddap(self):
        """
        Configure the ERDDAP dataset for the new sensor
        """
        rich.print(f"[purple]Running test {inspect.currentframe().f_code.co_name}")
        # def erddap_config(file: str, dataset_id: str, source_path: str, output: str = "", datasets_xml: str = ""):
        data_path = os.path.join("/erddapData", self.dataset_id)
        erddap_config(self.two_ctds_dataset,
                      "2CTDs",
                      data_path,
                      datasets_xml_file=self.datasets_xml)

        rich.print("Creating a hardFlag to force reload")
        dataset_hard_flag = os.path.join("erddapData", "hardFlag", self.dataset_id)
        with open(dataset_hard_flag, "w") as f:
            f.write("1")

        rich.print("now wait of erddap to process this flag...")

        while os.path.exists(dataset_hard_flag):
            time.sleep(1)
            rich.print("waiting for erddap to load the dataset...")

        time.sleep(3)
        dataset_url = self.erddap_url + "/tabledap/" + self.dataset_id + ".html"

        rich.print(dataset_url)
        urllib.request.urlretrieve(dataset_url)
        rich.print("[green]Dataset downloaded!")

        # now try to acess the data
        dataset_url = self.erddap_url + "/tabledap/" + self.dataset_id + ".nc"
        nc_file = "test.nc"
        urllib.request.urlretrieve(dataset_url, nc_file)
        wf = load_data(nc_file)
        df = wf.data



    @classmethod
    def tearDownClass(cls):
        rich.print("clearing datasets.xml backups...")
        files = os.listdir()
        for f in files:
            if f.startswith(".datasets.xml."):
                os.remove(f)
        rich.print("[green]ok")

        if False:
            os.system("docker compose down")
            rich.print("Clearing ERDDAP volume...")
            os.system("rm -rf erddapData/*")
            rich.print(f"[green]done")


if __name__ == "__main__":
    unittest.main(failfast=True, verbosity=1)


#!/usr/bin/env python3
"""
This class implements tools to generate and manage minimal metadata templates. Minimal metadata templates is the
minimum set of info required to generate a dataset.

author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 29/4/23
"""
from . import EmsoMetadata
from .autofill import autofill_minmeta
from .metadata_templates import global_metadata, sensor_metadata, variable_metadata, user_selectable_attributes, \
    choose_interactively
import rich
import os
from .dataset import get_qc_variables, get_variables, extract_netcdf_metadata
import json
import numpy as np

from .utils import avoid_filename_collision
from .waterframe import WaterFrame

def generate_min_meta_template(wf: WaterFrame, folder: str):
    """
    Takes a data frame and generates the a minimal metadata file from which the rest of the metadata can be derived
    """
    os.makedirs(folder, exist_ok=True)
    mfiles = []  # metadata files

    variables = get_variables(wf)
    m = {
        "README": {
            "*attr": "Mandatory attributes, must be set",
            "~attr": "Optional attributes, if not set will be guessed by the script",
            "$attr": "If not provided, will be requested interactively"
        },
        "global": global_metadata(),
        "variables": {},
        "sensor": sensor_metadata(),
        "coordinates": {
            "README": "If the dataset has fixed coordinates, please add them here as floats",
            "depth": "",
            "latitude": "",
            "longitude": ""
        }
    }

    for var in variables:
        m["variables"][var] = variable_metadata()

    datafile = wf.metadata["$datafile"]
    # create a filename
    a = os.path.basename(datafile).split(".")
    filename = ".".join(a[:-1]) + ".min.json"
    filename = os.path.join(folder, filename)

    # Do not overwrite file.txt, but create file(1).txt

    if os.path.exists(filename):
        filename = avoid_filename_collision(filename)

    with open(filename, "w") as f:
        f.write(json.dumps(m, indent=2))

    mfiles.append(filename)
    [rich.print(f"    {f}") for f in mfiles]


def process_selectable_metadata(m, filename=""):
    """
    Asks the user to interactively choose missing parameters
    """
    # Processing interactive values
    for key, value in m.copy().items():
        if key.startswith("$"):
            k = key[1:]
            if not value:
                if filename:
                    value = choose_interactively(k, filename, user_selectable_attributes()[k])
                else:
                    raise ValueError("Can't ask interactively for data if no filename is passed")
            m[key] = value  # remove leading $ and add user selected value
    return m


def remove_minmeta_keys(m):
    """
    Removes the minmeta leading keys (* ~ $)
    """
    # Process the rest of the params...
    for key, value in m.copy().items():
        if key.startswith("~") or key.startswith("*") or key.startswith("$"):
            del m[key]
            m[key[1:]] = value  # remove leading *

    if "README" in m.keys():
        del m["README"]
    return m


def load_full_meta(wf: WaterFrame, filename: str):
    """
    Loads a full metadata file
    """
    wf.metadata["$fullmeta"] = filename
    with open(filename) as f:
        metadata = json.load(f)

    sensor_ids = []
    for varname, varmeta in metadata["variables"].items():
        if "sensor_serial_number" in varmeta.keys():
            sensor_ids.append(varmeta['sensor_serial_number'])

    n = len(np.unique(sensor_ids))
    if n > 1:  # check if we have more than one serial number
        raise ValueError("Loading metadata a multisensor Dataset not yet implemented!!")
    elif n < 1:
        raise ValueError("sensor_serial_number not found!")

    wf.metadata["$sensor_id"] = sensor_ids[0]  # only one value, so get the first one
    return metadata


def load_min_meta(wf: WaterFrame, metadata: dict, emso: EmsoMetadata) -> dict:
    """
    Loads a minimal metadata file.
    """
    assert type(metadata) is dict, f"expected dict, got {type(metadata)}"
    minimal_metadata_file = ""
    if "$minmeta" in wf.metadata.keys():
        minimal_metadata_file = wf.metadata["$minmeta"]

    metadata["global"] = process_selectable_metadata(metadata["global"], filename=minimal_metadata_file)
    metadata["sensor"] = process_selectable_metadata(metadata["sensor"], filename=minimal_metadata_file)
    metadata["coordinates"] = process_selectable_metadata(metadata["coordinates"], filename=minimal_metadata_file)
    for var, m in metadata["variables"].items():
        metadata["variables"][var] = process_selectable_metadata(m, filename=minimal_metadata_file)

    # Make sure that we have all the necessary info
    check_mandatory_fields(metadata["global"])
    check_mandatory_fields(metadata["sensor"])
    for var, m in metadata["variables"].items():
        check_mandatory_fields(m)

    metadata = metadata.copy()
    metadata = autofill_minmeta(metadata, emso)

    if minimal_metadata_file:
        rich.print(f"Updating file {minimal_metadata_file} with selected user choices...", end="")
        with open(minimal_metadata_file, "w") as f:
            json.dump(metadata, f, indent=2)  # update the file, so
        rich.print("[green]done!")

    # Remove the leading keys
    metadata["global"] = remove_minmeta_keys(metadata["global"])
    metadata["sensor"] = remove_minmeta_keys(metadata["sensor"])
    metadata["coordinates"] = remove_minmeta_keys(metadata["coordinates"])

    for var, m in metadata["variables"].items():
        metadata["variables"][var] = remove_minmeta_keys(m)

    if "README" in metadata.keys():
        del metadata["README"]
    return metadata


def check_mandatory_fields(m):
    """
    Checks that all fields starting with * are filled
    """
    m = m.copy()
    error = False

    # First check all mandatory fields
    for key, value in m.copy().items():
        if key.startswith("*") and not value:
            error = True
            rich.print(f"[red]Mandatory field missing: \"{key}\"")
        if error:
            raise SyntaxError("Missing fields detected! Please fill all fields starting with '*'")

def np_encoder(object):
    """
    Encodes Numpy data for JSON lib
    """
    if isinstance(object, np.generic):
        return object.item()

def generate_full_metadata(wf: WaterFrame, folder):
    """
    Takes a waterframe and stores its full metadtaa in to a JSON file
    """

    os.makedirs(folder, exist_ok=True)
    # metadata file will be the datafile with full.json extension
    metafile = ".".join(wf.metadata["$datafile"].split(".")[:-1]) + ".full.json"
    metafile = os.path.join(folder, metafile)
    wf.metadata["$fullmeta"] = metafile
    rich.print(f"Storing full metadata into {metafile}...", end="")
    metadata = extract_netcdf_metadata(wf)
    with open(metafile, "w") as f:
        f.write(json.dumps(metadata, indent=2, default=np_encoder))
    rich.print("[green]done!")




#!/usr/bin/env python3
"""
Generates a chunk for the datasets.xml file

author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 28/4/23
"""
import os
import shutil

import lxml.etree as etree
from ..metadata.waterframe import WaterFrame
from ..metadata.dataset import get_variables, set_multisensor, get_dimensions, get_qc_variables
from ..metadata.xmlutils import get_element
from datetime import datetime
import rich


def generate_erddap_dataset(wf: WaterFrame, directory, dataset_id):
    """
    Generates a XML chunk to add it into ERDDAP's datasets.xml
    :param wf: waterframe with data and metadata
    :param directory: path where the NetCDF files will be stored
    :param dataset_id: datsetID to identify the dataset
    returns: a string containing the datasets.xml chunk to setup the dataset
    """
    dimensions = ["TIME", "LATITUDE", "LONGITUDE", "DEPTH"]  # custom dimensional order
    qc_variables = get_qc_variables(wf)

    # ERDDAP will force dimensions to be lowercase, so let's create a dict with source dest like:
    #     { "TIME": "time" }
    erddap_dims = {dim: dim.lower() for dim in dimensions}

    # To ensure that quality control variables match lowercase dimensions another dict like:
    #  {"LATITUDE_QC": "latitude_QC"}
    erddap_qc = {}
    for qcvar in qc_variables:
        source = qcvar.replace("_QC", "")
        if source in dimensions:
            erddap_qc[qcvar] = source.lower() + "_QC"  # ensure dimension is in lower case
        else:
            erddap_qc[qcvar] = qcvar  # do not modify

    # subset variables are QC vars and sensor_id
    subset_vars_str = ", ".join(erddap_qc.values())

    if "$multisensor" not in wf.metadata.keys():
        wf = set_multisensor(wf)

    if wf.metadata["$multisensor"]:
        erddap_dims["SENSOR_ID"] = "sensor_id"
        subset_vars_str += ", sensor_id"  # manually add as subset variable


    if "infoUrl" in wf.metadata.keys(): # If infoURL not set, use the edmo uri
        info_url = wf.metadata["infoUrl"]
    else:
        info_url = wf.metadata["institution_edmo_uri"]


    x = f"""
<dataset type="EDDTableFromMultidimNcFiles" datasetID="{dataset_id}" active="true">
    <reloadEveryNMinutes>10080</reloadEveryNMinutes>
    <updateEveryNMillis>10000</updateEveryNMillis>
    <fileDir>{directory}</fileDir>
    <fileNameRegex>.*</fileNameRegex>
    <recursive>true</recursive>    
    <pathRegex>.*</pathRegex>
    <metadataFrom>last</metadataFrom>
    <standardizeWhat>0</standardizeWhat>
    <removeMVRows>true</removeMVRows>
    <sortFilesBySourceNames></sortFilesBySourceNames>
    <fileTableInMemory>false</fileTableInMemory>
    <addAttributes>
        <att name="_NCProperties">null</att>
        <att name="cdm_data_type">Point</att>
        <att name="infoUrl">{info_url}</att>                
        <att name="sourceUrl">(local files)</att>
        <att name="standard_name_vocabulary">CF Standard Name Table v70</att>
        <att name="subsetVariables">{subset_vars_str}</att> 
    </addAttributes>        
</dataset>
    """
    tree = etree.ElementTree(etree.fromstring(x))
    root = tree.getroot()

    for source, dest in erddap_dims.items():  # already in lowercase
        datatype = "float"
        attrs = {}
        if dest == "time":
            datatype = "double"
            attrs = {
                "units": "seconds since 1970-01-01",
                "time_precision": "1970-01-01T00:00:00Z"
            }
        elif dest == "depth":
            attrs = {
                "units": "m",
            }
        elif dest == "sensor_id":
            datatype = "String"
        add_variable(root, source, dest, datatype, attributes=attrs)

    # Process all data variables
    for v in get_variables(wf):
        add_variable(root, v, v, "float", attributes={})

    for source, dest in erddap_qc.items():
        add_variable(root, source, dest, "byte", attributes={})
    etree.indent(root, space="    ", level=0)  # force indentation

    return serialize(tree)


def read_xml(filename):
    """
    Reads a XML file and returns the root element
    """
    parser = etree.XMLParser(remove_blank_text=True)
    tree = etree.parse(filename, parser)  # load from template
    root = tree.getroot()
    return root


def serialize(tree):
    etree.indent(tree, space="  ", level=0)
    return etree.tostring(tree, encoding="unicode", pretty_print=True, xml_declaration=False)


def prettyprint_xml(x):
    """
    produces a pretty print of an etree
    """
    etree.indent(x, space="  ", level=1)


def add_variable(root, source, destination, datatype, attributes: dict = {}):

    """
    Adds a variable to an ERDDAP dataset
    """

    __valid_data_types = ["int", "byte", "double", "float", "String"]

    if datatype not in __valid_data_types:
        raise ValueError(f"Data type '{datatype}' not valid!")

    var = etree.SubElement(root, "dataVariable")
    etree.SubElement(var, "sourceName").text = source
    etree.SubElement(var, "destinationName").text = destination
    etree.SubElement(var, "dataType").text = datatype
    attrs = etree.SubElement(var, "addAttributes")
    for key, value in attributes.items():
        att = etree.SubElement(attrs, "att")
        att.attrib["name"] = key
        att.text = value


def backup_datsets_file(filename):
    """
    Generates a .datasets.xml.YYYMMDD_HHMMSS backup file of the datasets.xml
    """
    assert type(filename) is str, f"expected string, got {type(filename)}"
    basename = os.path.basename(filename)
    directory = os.path.dirname(filename)
    backup = "." + basename + "." + datetime.now().strftime("%Y%m%d_%H%M%S")
    backup = os.path.join(directory, backup)
    shutil.copy2(filename, backup)
    return backup


def add_dataset(filename: str, dataset: str):
    """
    Adds a dataset to an exsiting ERDDAP deployment by modifying the datasets.xml config file
    :param filename: path to datasets.xml file
    :param dataset: string containing the XML configuration for the dataset
    """

    assert type(filename) is str, f"expected string, got {type(filename)}"
    assert type(dataset) is str, f"expected string, got {type(dataset)}"

    bckp = backup_datsets_file(filename)
    dataset_tree = etree.ElementTree(etree.fromstring(dataset))
    dataset_root = dataset_tree.getroot()
    dataset_id = dataset_root.attrib["datasetID"]

    parser = etree.XMLParser(remove_blank_text=True)
    tree = etree.parse(filename, parser)  # load from template
    root = tree.getroot()

    try:
        e = get_element(root, "dataset", attr="datasetID", attr_value=dataset_id)
        rich.print(f"[yellow]Overwriting existing dataset {dataset_id}!")
        e.getparent().remove(e)  # Remove the old dataset
    except LookupError:
        pass

    root.append(dataset_root)
    with open(filename, "w") as f:
        xml = etree.tostring(tree, encoding="UTF-8", pretty_print=True, xml_declaration=True)
        s = xml.decode()
        f.write(s)
















#!/usr/bin/env python3
"""
This file implements the tests to ensure that a dataset is harmonized. To include a new tests simply add a new method
like
 my_new_test(self, value, args) -> (bool, str)

author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 1/3/23
"""

import rich
from rich.console import Console
from rich.table import Table
from rich.style import Style
from rich.progress import Progress
import pandas as pd
import re
from . import EmsoMetadata
from .utils import group_metadata_variables
import inspect
import numpy as np

class EmsoMetadataTester:
    def __init__(self):
        """
        This class implements the tests to ensure that the metadata in a particular ERDDAP is harmonized with the EMSO
        metadata standards. The tests are configured in the 'EMSO_metadata.md' document. There should be 2 different
        tables with the tests defined, one for the global attributes and another one for tests to be carreid
        """
        # Dict to store all erddap. KEY is the test identifier while value is the method
        rich.print("[blue]Setting up EMSO Metadata Tests...")

        self.metadata = EmsoMetadata(True)

        self.implemented_tests = {}
        for name, element in inspect.getmembers(self):
            if name == "validate_dataset":  # exclude validate dataset from tests
                continue
            # Assume that all methods not starting with "_" are tests!
            elif inspect.ismethod(element) and not name.startswith("_"):
                self.implemented_tests[name] = element

        rich.print(f"Currently there are {len(self.implemented_tests)} tests implemented")

        # ensure that all required test are implemented...
        all_tests = list(self.metadata.global_attr["Compliance test"].values)
        all_tests += list(self.metadata.variable_attr["Compliance test"].values)
        all_tests = [value.split("#")[0] for value in all_tests]
        all_tests = np.unique(all_tests)
        error = False
        for test in all_tests:
            if test not in self.implemented_tests.keys():
                rich.print(f"[red]ERROR test {test} not implemented!")
                error = True
        if error:
            raise ValueError("Some tests are not implemented")

    def __process_results(self, df, verbose=False) -> (float, float, float):
        """
        Prints the results in a nice-looking table using rich
        :param df: DataFrame with test results
        """
        table = Table(title="Dataset Test Report")
        table.add_column("variable", justify="right", no_wrap=True, style="cyan")
        table.add_column("attribute", justify="right")
        table.add_column("required", justify="right")
        table.add_column("passed", justify="right")
        table.add_column("message", justify="right")
        table.add_column("value", justify="left")
        section = "global"
        for _, row in df.iterrows():
            # Process styles depending on the passed and required fields
            style = ""
            if row["message"] == "unimplemented":
                style = Style(color="bright_black", bold=False)
            elif row["message"] == "not defined":
                style = Style(color="medium_purple4", bold=False)
            elif row["required"] and not row["passed"]:
                style = Style(color="red", bold=True)
            elif row["passed"]:
                style = Style(color="green", bold=True)
            elif not row["required"] and not row["passed"]:
                style = Style(color="yellow", bold=False)

            if row["variable"] != section:  # add a new empty row with end section
                section = row["variable"]
                table.add_row(style=style, end_section=True)

            variable = row["variable"]
            attribute = row["attribute"]
            required = str(row["required"])
            passed = str(row["passed"])
            message = row["message"]
            value = str(row["value"])
            table.add_row(variable, attribute, required, passed, message, value, style=style)

        console = Console()
        console.print(table)

        df["required"] = df["required"].astype(bool)
        df["passed"] = df["passed"].astype(bool)

        r = df[df["required"]]  # required test
        req_tests = len(r)
        req_passed = len(r[r["passed"]])

        o = df[df["required"] == False]  # required test
        opt_tests = len(o)
        opt_passed = len(o[o["passed"]])

        total_tests = len(df)
        total_passed = len(df[df["passed"]])
        rich.print(f"Required tests passed: {req_passed} of {req_tests}")
        rich.print(f"Required tests passed: {opt_passed} of {opt_tests}")
        rich.print(f"   [bold]Total tests passed: {total_passed} of {total_tests}")

        def generate_bar_col(n):
            if n > 0.95:
                return "green"
            if n > 0.8:
                return "blue"
            if n > 0.6:
                return "yellow"
            if n > 0.4:
                return "dark_orange"
            return "red"

        t_color = generate_bar_col(total_passed / total_tests)
        r_color = generate_bar_col(req_passed / req_tests)
        o_color = generate_bar_col(opt_passed / opt_tests)

        with Progress(auto_refresh=False) as progress:
            req_task = progress.add_task(f"[{t_color}]Required tests...", total=req_tests)
            opt_task = progress.add_task(f"[{r_color}]Optional tests...", total=opt_tests)
            total_task = progress.add_task(f"[{o_color}]Total tests...", total=total_tests)

            progress.update(req_task, advance=req_passed)
            progress.update(opt_task, advance=opt_passed)
            progress.update(total_task, advance=total_passed)
            progress.stop()
        total = 100*round(total_passed / total_tests, 2)
        required = 100*round(req_passed / req_tests, 2)
        optional = 100*round(opt_passed / opt_tests, 2)

        return total, required, optional

    def __run_test(self, test_name, args, attribute: str, metadata, required, multiple, varname, results) -> (
    bool, str, any):
        """
        Applies the method test to the dict data and stores the output into results
        :param test_name: Name of the test tp apply
        :param args: arguments to be passed to test_method
        :param attribute: name of the attribute to be tested
        :param metadata: dict with the metadata being tested
        :param required: flag indicating if test is mandatory
        :param multiple: If set, multiple values separated by ; are allowed for this metadata field
        :param varname: variable name for the applied test (global for generic dataset metadata)
        :param results: dict with arrays to store the results of the tests
        :return: a tuple with (bool, str, any). Boolean indicates success, str is an error message and any is the value
                 of the attribute or None if not present.
        """

        if attribute == "$name" or attribute in metadata.keys():
            if test_name not in self.implemented_tests.keys():
                rich.print(f"[red]Test '{test_name}' not implemented!")
                raise LookupError(f"Test {test_name} not found")

            if attribute == "$name":
                value = varname
            else:
                value = metadata[attribute]

            if type(value) == str and ";" in value:
                values = value.split(";")  # split multiple values
                for i in range(len(values)):
                    if values[i].startswith(" "):
                        values[i] = values[i][1:]  # Make sure that first space is not kept
            else:
                values = [value]

            if len(values) > 1 and not multiple:
                # got multiple values for a single-value test!
                passed = False
                message = f"Got {len(values)} values for a single-value test"
            else:
                messages = []
                passed_flags = []
                for v in values:
                    test_method = self.implemented_tests[test_name]
                    try:
                        p, m = test_method(v, args)  # apply test method
                    except Exception as e:
                        rich.print(
                            f"[red]Error when executing test '{test_name}' with arguments '{args}' and value '{v}'")
                        raise e
                    if not m:
                        m = "ok"  # instead of empty message just leave ok

                    messages.append(m)
                    passed_flags.append(p)
                message = "; ".join(messages)
                passed = True

                for p in passed_flags:
                    passed = p and passed
        else:  # Not found
            passed = False
            message = "not found"
            value = ""

        results["attribute"].append(attribute)
        results["variable"].append(varname)
        results["passed"].append(passed)
        results["required"].append(required)
        results["message"].append(message)
        results["value"].append(value)
        return passed, message, value

    def __test_group_handler(self, test_group: pd.DataFrame, metadata: dict, variable: str, verbose: bool, results={}
                             ) -> dict:
        """
        Takes a list of tests from the metadata specification and applies it to the metadata json structure
        :param test_group: DataFrame of the group of tests required
        :param metadata: JSON structure (dict) under test
        :param variable: Variable being tested, 'global' for global dataset attributes
        :param verbose: if True, will add attributes present in the dataset but not required by the standard.
        :param results: a dict to store the results. If empty a new one will be created
        :returns: result structure
        """
        if not results:
            results = {
                "attribute": [],
                "variable": [],
                "required": [],
                "passed": [],
                "message": [],
                "value": []
            }
        attribute_col = test_group.columns[0]

        # Run Global Attributes test
        for _, row in test_group.iterrows():
            attribute = row[attribute_col]
            test_name = row["Compliance test"]
            required = row["Required"]
            multiple = row["Multiple"]
            if not test_name:
                rich.print(f"[yellow]WARNING: test for {attribute} not implemented!")
                continue

            args = []
            if "#" in test_name:
                test_name, args = test_name.split("#")
                args = args.split(",")  # comma-separated fields are args

            self.__run_test(test_name, args, attribute, metadata, required, multiple, variable, results)

        if verbose:  # add all parameters not listed in the standard
            checks = list(test_group[attribute_col].values)
            for key, value in metadata.items():
                if key not in checks:
                    results["attribute"].append(key)
                    results["variable"].append(variable)
                    results["passed"].append("n/a")
                    results["required"].append("n/a")
                    results["message"].append("not defined")

                    if type(value) == str and len(value) > 100:
                        value = value.strip()[:60] + "..."
                    results["value"].append(value)
        return results

    def validate_dataset(self, metadata, verbose=True, store_results=False):
        """
        Takes the well-formatted JSON metadata from an ERDDAP dataset and processes it
        :param metadata: well-formatted JSON metadta for an ERDDAP dataset
        :return: a DataFrame with the following columns: [attribute, variable, required, passed, message, value]
        """

        metadata = group_metadata_variables(metadata)

        # Try to get a dataset id
        if "dataset_id" in metadata["global"].keys():
            dataset_id = metadata["global"]["dataset_id"]
        elif "id" in metadata["global"].keys():
            dataset_id = metadata["global"]["id"]
        else:
            dataset_id = metadata["global"]["title"]

        rich.print(f"#### Validating dataset [cyan]{metadata['global']['title']}[/cyan] ####")

        # Test global attributes
        results = self.__test_group_handler(self.metadata.global_attr, metadata["global"], "global", verbose)

        # Test every dimension
        for varname, var_metadata in metadata["dimensions"].items():
            if varname.lower() == "sensor_id":
                # Deliberately skip sensor_id
                continue
            # First check variable name manually

            results = self.__test_group_handler(self.metadata.dimension_attr, metadata["dimensions"][varname], varname,
                                                verbose, results)

        # Test every variable
        for varname, var_metadata in metadata["variables"].items():
            results = self.__test_group_handler(self.metadata.variable_attr, metadata["variables"][varname], varname,
                                                verbose, results)
        # Test every QC column
        for varname, var_metadata in metadata["qc"].items():
            results = self.__test_group_handler(self.metadata.qc_attr, metadata["qc"][varname], varname,
                                                verbose, results)

        for varname, var_metadata in metadata["technical"].items():
            results = self.__test_group_handler(self.metadata.technical_attr, metadata["technical"][varname], varname,
                                                verbose, results)

        df = pd.DataFrame(results)
        total, required, optional = self.__process_results(df, verbose=verbose)
        r = {
            "dataset_id": dataset_id,
            "institution": "unknown",
            "emso_facility": "",
            "total": total,
            "required": required,
            "optional": optional
        }

        if "institution" in metadata["global"].keys():
            r["institution"] = metadata["global"]["institution"]
        elif "institution_edmo_codi" in metadata["global"].keys():
            r["institution"] = "EMDO Code " + metadata["global"]["institution_edmo_codi"]
        else:
            r["institution"] = "unknown"

        # Add EMSO Facility in results
        if "emso_facility" in metadata["global"].keys():
            r["emso_facility"] = metadata["global"]["emso_facility"]

        if store_results:
            results_csv = f"report_{dataset_id}.csv".replace(" ", "_").replace(",", "")
            rich.print(f"[green]Storing results into file {results_csv}...")
            df.to_csv(results_csv, index=False)
        return r

    # ------------------------------------------------ TEST METHODS -------------------------------------------------- #
    # Test methods implement checks to be applied to a group metadata attributes, such as coordinates or valid email.
    # All tests should return a tuple (bool, str) tuple. The bool indicates success (true/false), while the message str
    # indicates in plain text the reason why the test failed. If the test successfully passes sucess, the return str
    # should be empty.
    # ---------------------------------------------------------------------------------------------------------------- #

    # ------------ EDMO -------- #
    def edmo_code(self, value, args):
        if type(value) == str:
            rich.print("[yellow]WARNING: EDMO code should be integer! converting from string to int")
            try:
                value = int(value)
            except ValueError:
                return False, f"'{value}' is not a valid EDMO code"
        if value in self.metadata.edmo_codes["code"].values:
            return True, ""
        return False, f"'{value}' is not a valid EDMO code"

    def edmo_uri(self, value, args):
        if type(value) != str:
            return False, "EDMO URI should be a string"

        uri = value.replace("http", "https")  # make sure to use http
        if uri.endswith("/"):
            uri = uri[:-1]  # remove ending /


        if value in self.metadata.edmo_codes["uri"].values:
            return True, ""

        return False, f"'{value}' is not a valid EDMO code"

    # -------- SeaDataNet -------- #
    def sdn_vocab_urn(self, value, args):
        """
        Tests that the value is a valid URN for a specific SeaDataNet Vocabulary. the vocab should be specified in *args
        """
        if len(args) != 1:
            raise SyntaxError("Vocabulary identifier should be passed in args, e.g. 'P01'")
        vocab = args[0]

        if vocab not in self.metadata.sdn_vocabs_ids.keys():
            raise ValueError(
                f"Vocabulary '{vocab}' not loaded! Loaded vocabs are {self.metadata.sdn_vocabs_ids.keys()}")

        if value in self.metadata.sdn_vocabs_ids[vocab]:
            return True, ""

        return False, f"Not a valid '{vocab}' URN"

    def sdn_vocab_pref_label(self, value, args):
        """
        Tests that the value is a valid prefered label for a SeaDataNet Vocabulary. the vocab should be specified in
        *args
        """
        if len(args) != 1:
            raise SyntaxError("Vocabulary identifier should be passed in args, e.g. 'P01'")
        vocab = args[0]
        if vocab not in self.metadata.sdn_vocabs_pref_label.keys():
            raise ValueError(
                f"Vocabulary '{vocab}' not loaded! Loaded vocabs are {self.metadata.sdn_vocabs_pref_label.keys()}")

        if value in self.metadata.sdn_vocabs_pref_label[vocab]:
            return True, ""

        return False, f"Not a valid '{vocab}' prefered label"

    def cf_standard_name(self, value, args):
        """
        Tests that the value is a valid prefered label for a SeaDataNet Vocabulary. the vocab should be specified in
        *args
        """
        vocab = "P07"
        if vocab not in self.metadata.sdn_vocabs_pref_label.keys():
            raise ValueError(
                f"Vocabulary '{vocab}' not loaded! Loaded vocabs are {self.metadata.sdn_vocabs_pref_label.keys()}")

        if value in self.metadata.sdn_vocabs_pref_label[vocab]:
            return True, ""
        return False, f"Not a valid '{vocab}' prefered label"

    def sdn_vocab_uri(self, value, args):
        """
        Tests that the value is a valid URI for a SeaDataNet Vocabulary. the vocab should be specified in
        *args
        """
        if len(args) != 1:
            raise SyntaxError("Vocabulary identifier should be passed in args, e.g. 'P01'")
        vocab = args[0]

        uri = value.replace("https", "http")  # make sure to use http

        if not uri.endswith("/"):
            uri += "/"  # make sure that the uri ends with /

        if vocab not in self.metadata.sdn_vocabs_uris.keys():
            raise ValueError(
                f"Vocabulary '{vocab}' not loaded! Loaded vocabs are {self.metadata.sdn_vocabs_uris.keys()}")

        if uri in self.metadata.sdn_vocabs_uris[vocab]:
            return True, ""

        return False, f"Not a valid '{vocab}' URI"

    # --------- OceanSITES -------- #
    def oceansites_sensor_mount(self, value, args):
        if value in self.metadata.oceansites_sensor_mount:
            return True, ""
        return False, f"Sensor mount not valid, valid values are {self.metadata.oceansites_sensor_mount}"

    def oceansites_sensor_orientation(self, value, args):
        if value in self.metadata.oceansites_sensor_orientation:
            return True, ""
        return False, f"Sensor orientation not valid, valid values are {self.metadata.oceansites_sensor_orientation}"

    def oceansites_data_type(self, value, args):
        if value in self.metadata.oceansites_data_types:
            return True, ""
        return False, f"Data type not valid, valid values are {self.metadata.oceansites_data_types}"

    def oceansites_data_mode(self, value, args):
        if value in self.metadata.oceansites_data_modes:
            return True, ""
        return False, f"Data mode not valid, valid values are {self.metadata.oceansites_data_modes}"

    # -------- EMSO Data -------- #
    def emso_facility(self, value, args):
        if value in self.metadata.emso_regional_facilities:
            return True, ""
        return False, f"not a valid EMSO Regional Facility"

    def emso_site_code(self, value, args):
        if value in self.metadata.emso_sites:
            return True, ""
        return False, f"not a valid EMSO site"

    # -------- SPDX Licenses -------- #
    def spdx_license_name(self, value, args):
        if value in self.metadata.spdx_license_names:
            return True, ""
        return False, "Not a valid SPDX license code"

    def spdx_license_uri(self, value, args):
        value = value.replace("http://", "https://")  # ensure https
        value = value.replace(".jsonld", "").replace(".json", "").replace(".html", "")  # delete format
        if value in self.metadata.spdx_license_uris.values():
            return True, ""
        return False, f"Not a valid SDPX license uri '{value}'"

    # -------- Geospatial Coordinates -------- #
    def coordinate(self, value, args) -> (bool, str):
        """
        Checks if a coordinate is valid. Within args a single string indicating "latitude" "longitude" or "depth" must
        be passed
        """
        __cordinate_types = ["latitude", "longitude", "depth"]
        if len(args) != 1:
            raise SyntaxError("Coordinate type should be passed in args, e.g. 'P01'")

        coordinate = args[0].lower()  # force lowercase
        if coordinate not in __cordinate_types:
            raise SyntaxError(f"Coordinate type should be 'latitude', 'longitude' or 'depth'")
        try:
            value = float(value)
        except ValueError:
            return False, f"Could not convert '{value}' to float"

        if coordinate == "latitude" and (value < -90 or value > 90):
            return False, "latitude should be between -90 and +90"
        elif coordinate == "longitude" and (value < -180 or value > 180):
            return False, "longitude should be between -90 and +90"
        # depth is valid from a 2km tall mountain to the depths of the mariana trench
        elif coordinate == "depth" and (value < -2000 or value > 11000):
            return False, "depth should be between -2000 and 11000 metres"

        return True, ""

    # --------- Other tests -------- #
    def equals(self, value, args):
        if value == args[0]:
            return True, ""
        return False, f"expected value {args[0]}"

    def data_type(self, value, args) -> (bool, str):
        """
        Check if value is of the exepcted type.
        :param value: value to be tested
        :param args: list with one value containing a string of the type, like ['string'] or ['float']
        :returns: passed, error message
        """
        if len(args) != 1:
            raise ValueError("Expected exacly one extra argument with type")
        data_type = args[0]

        if data_type in ["str", "string"]:
            # check string
            if type(value) != str:
                return False, "not a string"

        elif data_type in ["int", "integer", "unsigned"]:
            # check string
            if type(value) != int:
                return False, "not an integer"

        elif data_type in ["float", "double"]:
            # check string
            if type(value) != float:
                return False, "not a float"

        elif data_type in ["date"]:
            return False, "unimplemented"

        elif data_type in ["datetime"]:
            try:
                pd.Timestamp(value)
            except ValueError:
                return False, "Datetime not valid, expecting format 'YYY-dd-mmTHH:MM:SS+tz'"
        else:
            raise ValueError(f"Unrecodgnized data type '{data_type}'...")

        return True, ""

    def email(self, value, args) -> (bool, str):
        if len(value) > 7:
            if re.match("^.+@(\[?)[a-zA-Z0-9-.]+.([a-zA-Z]{2,3}|[0-9]{1,3})(]?)$", value):
                return True, ""
        return False, f"email '{value}' not valid"

    def valid_doi(self, value, args) -> (bool, str):
        if re.match(r"^10.\d{4,9}/[-._;()/:A-Za-z0-9]+$", value):
            return True, ""
        return False, f"DOI '{value}' not valid"

    def check_variable_name(self, value, args) -> (bool, str):
        """
        Checks if a variable name exists in:
            1. OceanSITES
            2. P02
            3. Copernicus Params

        If not throw a warning
        """
        if value in self.metadata.oceansites_param_codes:
            return True, "Variable name found in OceanSITES"
        elif value in self.metadata.sdn_p02_names:
            return True, "Variable name found in P02"
        elif value in self.metadata.copernicus_variables:
            return True, "Variable name found in Copernicus INSTAC codes"
        else:
            return False, "Parameter name not found in OceanSITES, P02 and Copernicus!"




#!/usr/bin/env python3
"""
Miscellaneous functions

author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 26/4/23
"""
import rich
from rich.progress import Progress
import urllib
import concurrent.futures as futures
import os
from .constants import dimensions
import numpy as np


def group_metadata_variables(metadata):
    """
    Takes a dictionary with all the variables in the "variable" and groups them into "variables", "qualityControl" and
    "dimensions"
    """

    m = metadata.copy()

    vars = list(m["variables"].keys())

    qcs = {key: m["variables"].pop(key) for key in vars if key.upper().endswith("_QC")}
    stds = {key: m["variables"].pop(key) for key in vars if key.upper().endswith("_STD")}
    dims = {key: m["variables"].pop(key) for key in vars if key.upper() in dimensions}

    technical = {}
    for key in vars:
        try:
            if "technical_data" in m["variables"][key].keys():
                rich.print(f"variable {key} is 'technical'")
                technical[key] = m["variables"].pop(key)
        except KeyError:
            rich.print(f"variable {key} is 'scientific'")

    m = {
        "global": m["global"],
        "variables": m["variables"],
        "qc": qcs,
        "dimensions": dims,
        "std": stds,
        "technical": technical
    }
    return m


def __threadify_index_handler(index, handler, args):
    """
    This function adds the index to the return of the handler function. Useful to sort the results of a
    multi-threaded operation
    :param index: index to be returned
    :param handler: function handler to be called
    :param args: list with arguments of the function handler
    :return: tuple with (index, xxx) where xxx is whatever the handler function returned
    """
    result = handler(*args)  # call the handler
    return index, result  # add index to the result


def threadify(arg_list, handler, max_threads=10):
    """
    Splits a repetitive task into several threads
    :param arg_list: each element in the list will crate a thread and its contents passed to the handler
    :param handler: function to be invoked by every thread
    :param max_threads: Max threads to be launched at once
    :return: a list with the results (ordered as arg_list)
    """
    index = 0  # thread index
    with futures.ThreadPoolExecutor(max_workers=max_threads) as executor:
        threads = []  # empty thread list
        results = []  # empty list of thread results
        for args in arg_list:
            # submit tasks to the executor and append the tasks to the thread list
            threads.append(executor.submit(__threadify_index_handler, index, handler, args))
            index += 1

        # wait for all threads to end
        for future in futures.as_completed(threads):
            future_result = future.result()  # result of the handler
            results.append(future_result)

        # sort the results by the index added by __threadify_index_handler
        sorted_results = sorted(results, key=lambda a: a[0])

        final_results = []  # create a new array without indexes
        for result in sorted_results:
            final_results.append(result[1])
        return final_results


def download_file(url, file):
    """
    wrapper for urllib.error.HTTPError
    """
    try:
        return urllib.request.urlretrieve(url, file)
    except urllib.error.HTTPError as e:
        rich.print(f"[red]{str(e)}")
        rich.print(f"[red]Could not download from {url} to file {file}")
        raise e


def download_files(tasks, force_download=False):
    if len(tasks) == 1:
        return None
    args = []
    for url, file, name in tasks:
        if os.path.isfile(file) and not force_download:
            pass
        else:
            args.append((url, file))

    threadify(args, download_file)


def drop_duplicates(df, timestamp="time"):
    """
    useful for datasets that have duplicated values with consecutive timestamps (e.g. data is generated minutely, but
    inserted into a database every 20 secs). So the following dataframe:

                                col1      col2    col3
        timestamp
        2020-01-01 00:00:00    13.45    475.45    12.7
        2020-01-01 00:00:00    13.45    475.45    12.7
        2020-01-01 00:00:00    13.45    475.45    12.7
        2020-01-01 00:01:00    12.89    324.12    78.8
        2020-01-01 00:01:00    12.89    324.12    78.8
        2020-01-01 00:01:00    12.89    324.12    78.8
        ...

    will be simplified to:

                                col1      col2    col3
        timestamp
        2020-01-01 00:00:00    13.45    475.45    12.7
        2020-01-01 00:01:00    12.89    324.12    78.8

    :param df: input dataframe
    :return: simplified dataframe
    """
    if df.empty:
        rich.print("[yellow]WARNING empty dataframe")
        return df
    columns = [col for col in df.columns if col != timestamp]
    del_array = np.zeros(len(df))  # create an empty array
    duplicates = 0
    with Progress() as progress:  # Use Progress() to show a nice progress bar
        task = progress.add_task("Detecting duplicates", total=len(df))
        init = True
        for index, row in df.iterrows():
            progress.update(task, advance=1)
            if init:
                init = False
                last_valid_row = row
                continue

            diff = False  # flag to indicate if the current column is different from the last valid
            for column in columns:  # compare value by value
                if row[column] != last_valid_row[column]:
                    # column is different
                    last_valid_row = row
                    diff = True

                    break
            if not diff:  # there's no difference between columns, so this one needs to be deleted
                del_array[duplicates] = index
                duplicates += 1

    print(f"Duplicated lines {duplicates} from {len(df)}, ({100*duplicates/len(df):.02f} %)")
    del_array = del_array[:duplicates]  # keep only the part of the array that has been filled
    rich.print("dropping rows...")
    df.drop(del_array, inplace=True)
    return df


def avoid_filename_collision(filename):
    """
    Takes a filename (e.g. data.txt) and converts it to an available filename (e.g. data(1).txt).
    """
    i = 1
    a = filename.split(".")
    a[0] = a[0] + f"({i})"
    filename = ".".join(a)
    while os.path.isfile(filename):
        i += 1
        filename = filename.split("(")[0] + f"({i})" + filename.split(")")[1]
    return filename


def merge_dicts(strong: dict, weak: dict):
    """
    Merges two dictionaries. If a duplicated field is detected the 'strong' value will prevail
    """
    out = weak.copy()
    out.update(strong)
    return out


def get_file_list(dir_name):
    """
     create a list of file and sub directories names in the given directory
     :param dir_name: directory name
     :returns: list of all files with relative path
     """
    file_list = os.listdir(dir_name)
    all_files = list()
    for entry in file_list:
        full_path = os.path.join(dir_name, entry)
        if os.path.isdir(full_path):
            all_files = all_files + get_file_list(full_path)
        else:
            all_files.append(full_path)
    return all_files



#!/usr/bin/env python3

dimensions = ["TIME", "LATITUDE", "LONGITUDE", "DEPTH", "SENSOR_ID"]
iso_time_format = "%Y-%m-%dT%H:%M:%SZ"
qc_flags = {
    "unknown": 0,
    "good_data": 1,
    "probably_good_data": 2,
    "potentially_correctable_bad_data": 3,
    "bad_data": 4,
    "nominal_value": 7,
    "interpolated_value": 8,
    "missing_value": 9
}

fill_value = -999999  # default, for floats
fill_value_uint8 = 254


"""
Common functions and classes defined here

@author: Enoc Martínez
@contact: enoc.martinez@upc.edu
"""

import sys
import xml.etree as etree

# Namespaces used by SensorML, SOS and O&M Ddocuments


ns = {'swe':'http://www.opengis.net/swe/2.0',
      'sos': 'http://www.opengis.net/sos/2.0',
      'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
      'ns': 'http://www.opengis.net/swe/2.0',
      'xlink': 'http://www.w3.org/1999/xlink',
      'swes': 'http://www.opengis.net/swes/2.0',
      'gml': 'http://www.opengis.net/gml/3.2',
      'sml': 'http://www.opengis.net/sensorml/2.0',
      'om': 'http://www.opengis.net/om/2.0',
      'sams': 'http://www.opengis.net/samplingSpatial/2.0',
      'sf': 'http://www.opengis.net/sampling/2.0',
      'ows': 'http://www.opengis.net/ows/1.1',
      'drt': 'http://www.opengis.net/drt/1.0',
      'gda': 'http://www.opengis.net/sosgda/1.0',
      "gmd": "http://www.isotc211.org/2005/gmd",
      "gco": "http://www.isotc211.org/2005/gco"}

__enable_debug__ = False


def dbg(*arguments):
    state = False
    if __enable_debug__:
        s = str(*arguments)
        for char in s:
            if state == False and char == '\"':
                state = True
            elif state == True and char == '\"':
                state = False
            else:
                sys.stdout.write(char)
        sys.stdout.write('\n')


def has_namespace(tag):
    """This function checks if a tag has a namespace"""
    if ':' in tag:
        return True
    else:
        return False


def true_false_string(x):
    """Converts from boolean type to 'true' or 'false' string"""
    if type(x) != type(True):
        raise ValueError("Input not boolean")

    elif x == True:
        return 'true'
    else:
        return 'false'


class GenericXML:
    """
    Implements Generic functions for XML manipulation
    """

    def __init__(self):
        self.xml = ""
        pass

    def generate_document(self, outfile="", header=False):
        """
        Generates an XML document based on self.tree structure. If outfile is set the document is
        stored in a file.
        :param outfile: if set stores the XML document in file
        :return: xml document as string
        """
        self.xml = etree.tostring(self.tree, encoding="UTF-8", pretty_print=True, xml_declaration=True)
        s = self.xml.decode()
        if outfile:
            with open(outfile, 'w') as f:
                if header:
                    pass
                f.write(s)
        return self.xml


def set_child_label(root, newvalue):
    """Modifies the content of the children's label element"""
    return generic_set_child_element(root, newvalue, 'label')


def set_child_definition(root, newvalue):
    """Modifies the content of the children's definition element"""
    return generic_set_child_element(root, 'definition', newvalue)


def set_child_value(root, newvalue):
    """Modifies the content of the children's value element"""
    return generic_set_child_element(root, newvalue, 'value')


def set_value_to_child(root, newvalue, tag, attr=None, attr_value=None):
    """Looks for the an element, then looks for its ns:value child and sets the text value"""
    element = get_element(root, tag, attr=attr, attr_value=attr_value)
    set_child_value(element, newvalue)


def set_label_to_child(root, newvalue, tag, attr=None, attr_value=None):
    """Looks for the an element, then looks for its ns:value child and sets the text value"""
    element = get_element(root, tag, attr=attr, attr_value=attr_value)
    set_child_label(element, newvalue)


def generic_set_child_element(root, newvalue, childtag, attr=None, attr_value=None):
    """Generic method that modifies the value of a child element"""
    child = None
    if has_namespace(childtag) == True:
        child = get_element(newvalue, childtag, attr=attr, attr_value=attr_value)

    # Loop through all the namespaces if namespace not provided
    else:
        for currents in list(ns.keys()):
            tag_with_ns = currents + ':' + childtag
            try:
                child = get_element(root, tag_with_ns, attr, attr_value)
            except LookupError:
                pass
            if child != None:
                break
    if child == None:
        raise LookupError("Child element \"%s\" not found" % childtag)

    child.text = newvalue


def set_child_attribute(root, childtag, attr, attr_value, namespace=None):
    child = None
    if 'namespace' != None:
        xpath = './' + namespace + ':' + childtag
        child = root.find(xpath, namespace=ns)

    # Loop through all the namespaces if namespace not provided
    else:
        for currentns in list(ns.keys()):
            xpath = './' + currentns + ':' + childtag
            child = root.find(xpath, namespace=ns)
            if child != None:
                break

    if child == None:
        raise ValueError("Child element %s not found", childtag)

    print("Setting attr", attr, "to", attr_value)
    # Modify the dictionary containing the child's attributes
    if attr in child.attrib:
        child.attrib[attr] = attr_value
        # if element does not exist append a new element to the dictionary
    else:
        new_attribute = {attr: attr_value}
        child.attrib.update(new_attribute)  # Update the attribute dictionary


def set_value(root, value, tag, attr=None, attr_value=None):
    """Looks for an elements and sets it's value"""
    element = get_element(root, tag, attr=attr, attr_value=attr_value)
    element.text = value
    return element


def set_attribute(root, value, tag, attr):
    """Looks for an elements and sets it's value"""
    element = get_element(root, tag, attr=attr)
    element.attrib[attr] = value
    return element


def remove_element(root, tag, attr=None, attr_value=None):
    """
    looks for an element and removes it from the tree
    :param root: etree root to look
    :param tag: element tag (including namespace)
    :param attr: attribute name
    :param attr_value: attribute value
    """
    element = get_element(root, tag, attr=attr, attr_value=attr_value)
    element.getparent().remove(element)


def get_elements(root, tag, attr=None, attr_value=None):
    """
    Looks in root for all elements with matching tag, attribute name and attribute value
    :param root: etree root to look
    :param tag: element tag (including namespace)
    :param attr: attribute name
    :param attr_value: attribute value
    :returns: list of matching elements
    """
    dbg('looking for tag \"%s\" attr \"%s\" attr_value \"%s\"' % (tag, attr, attr_value))
    # If no attributes only one candidates should be found by tag, otherwise error
    if tag != None and attr == None:
        xpath = './/' + tag
        return root.findall(xpath, namespaces=ns)

    # Element with attribute, get all subelements that match the tag with the attribute
    xpath = './/' + tag + '[@%s]' % attr
    candidates = root.findall(xpath, namespaces=ns)
    dbg('got %d elements  with tag \"%s\"' % (len(candidates), tag))

    selected = []
    for candidate in candidates:
        if attr in candidate.attrib.keys():
            selected.append(candidate)

    dbg('got %d elements  with attr \"%s\"' % (len(selected), attr))

    # If attribute without value
    if attr_value == None:
        return selected

    candidates = selected

    # If tag, attribute and value multiple values can be found, loop until matching
    # attribute value is found
    selected = []
    for element in candidates:
        if element.attrib.get(attr) == attr_value:
            selected.append(element)

    if len(selected) > 0:
        dbg('got %d elements  with value \"%s\"' % (len(selected), attr_value))
        return selected

    raise LookupError("Element not found %s %s=%s" % (tag, attr, attr_value))


def get_element(root, tag, attr=None, attr_value=None):
    """
    Looks in root for an elements with matching tag, attribute name and attribute value.
    If more than one element is found, error is raised
    :param root: etree root to look
    :param tag: element tag (including namespace)
    :param attr: attribute name
    :param attr_value: attribute value
    :returns: list of matching elements
    """
    candidates = get_elements(root, tag, attr=attr, attr_value=attr_value)

    if len(candidates) == 1:
        return candidates[0]
    elif len(candidates) > 1:
        raise LookupError("Too many elements %s found, expected 1 got %d" % (tag, len(candidates)))

    raise LookupError("Element not found %s %s=%s" % (tag, attr, attr_value))


def get_children_index(root, tag):
    """
    This function looks at the children of root and returns the index of the children with matching tag
    :param root: etree element where to look for children
    :param tag: children tag
    :return: children index with matching tag
    :raises: LookupError if no matching children is found
    """
    if ":" in tag:
        # replace tag with namespace
        ns_tag = tag.split(':')[0]
        tag = "{%s}" % ns[ns_tag] + tag.split(":")[1]

    i = 0
    mylist = list(root)
    for i in range(0, len(mylist)):
        child = root[i]
        if child.tag == tag:
            return i
    raise LookupError("Children with tag %s not found in element %s " % (tag, root.tag))


def append_after(root, tag, element, comment=""):
    """
    This function appends "element" to the children list in root just after the children with matching tag
    :param root: etree element where to append
    :param tag: children tag after which the element will be appended
    :param element: etree Element to append
    :param comment: if comment is set, a comment with this text will be added before the element
    :raises: LookupError if no matching children is found
    """
    i = get_children_index(root, tag)
    if len(comment) > 0:
        comment_object = etree.Comment(comment)
        root.insert(i+1, comment_object)
        i += 1

    root.insert(i+1, element)


def append_before(root, tag, element, comment=""):
    """
    This function appends "element" to the children list in root just before the children with matching tag
    :param root: etree element where to append
    :param tag: children tag after which the element will be appended
    :param element: etree Element to append
    :param comment: if comment is set, a comment with this text will be added before the element
    :raises: LookupError if no matching children is found
    """
    i = get_children_index(root, tag)

    if len(comment) > 0:
        comment_object = etree.Comment(comment)
        root.insert(i, comment_object)
        i += 1

    root.insert(i, element)






#!/usr/bin/env python3
"""
Custom file to create NetCDF files


author: Enoc Martínez
institution: Universitat Politècnica de Catalunya (UPC)
email: enoc.martinez@upc.edu
license: MIT
created: 18/4/23
"""
import netCDF4 as nc
import pandas as pd
import numpy as np
from .constants import fill_value, fill_value_uint8
import xarray as xr
from .waterframe import WaterFrame


def wf_to_multidim_nc(wf: WaterFrame, filename: str, dimensions: list, fill_value=fill_value, time_key="TIME",
                      join_attr="; ", fill_value_uint8=fill_value_uint8):
    """
    Creates a multidimensinoal NetCDF-4 file
    :param filename: name of the output file
    :param df: pandas dataframe with the data
    :param metadata: dict containing metadata
    :param multiple_sensors
    """

    # Make sure that time is the last entry in the multiindex
    if time_key in dimensions:
        dimensions.remove(time_key)
        dimensions.append(time_key)

    df = wf.data  # Access the DataFrame within the waterframe
    index_df = df[dimensions].copy()  # create a dataframe with only the variables that will be used as indexes
    multiindex = pd.MultiIndex.from_frame(index_df)  # create a multiindex from the dataframe

    # Arrange other variables into a dict
    data = {col: df[col].values for col in df.columns if col not in dimensions}

    # Create a dataframe with multiindex
    data_df = pd.DataFrame(data, index=multiindex)
    dimensions = tuple(dimensions)

    with nc.Dataset(filename, "w", format="NETCDF4") as ncfile:
        for dimension in dimensions:
            data = index_df[dimension].values
            values = np.unique(data)  # fixed-length dimension
            if dimension == time_key:
                # convert timestamp to float
                index_df[time_key] = pd.to_datetime(index_df[time_key])
                times = index_df[time_key].dt.to_pydatetime()
                values = nc.date2num(times, "seconds since 1970-01-01", calendar="standard")

            ncfile.createDimension(dimension, len(values))  # create dimension
            if type(values[0]) == str:  # Some dimension may be a string (e.g. sensor_id)
                # zlib=False because variable-length strings cannot be compressed
                var = ncfile.createVariable(dimension, str, (dimension,), fill_value=fill_value, zlib=False)
            else:
                var = ncfile.createVariable(dimension, 'float', (dimension,), fill_value=fill_value, zlib=True)

            var[:] = values  # assign dimension values

            # add all dimension metadata

            for key, value in wf.vocabulary[dimension].items():
                if type(value) == list:
                    values = [str(v) for v in value]
                    value = join_attr.join(values)
                var.setncattr(key, value)

        for varname in data_df.columns:
            values = data_df[varname].to_numpy()  # assign values to the variable
            if varname.endswith("_QC"):
                # Store Quality Control as unsigned bytes
                var = ncfile.createVariable(varname, "u1", dimensions, fill_value=fill_value_uint8, zlib=True)
                var[:] = values.astype(np.int8)
            else:
                var = ncfile.createVariable(varname, 'float', dimensions, fill_value=fill_value, zlib=True)
                var[:] = values

            # Adding metadata
            for key, value in wf.vocabulary[varname].items():
                if type(value) == list:
                    values = [str(v) for v in value]
                    value = join_attr.join(values)
                var.setncattr(key, value)

        # Set global attibutes
        for key, value in wf.metadata.items():
            if type(value) == list:
                values = [str(v) for v in value]
                value = join_attr.join(values)
            ncfile.setncattr(key, value)


def read_nc(path, decode_times=True, time_key="TIME"):
    """
    Read data form NetCDF file and create a WaterFrame.

    Parameters
    ----------
        path: str
            Path of the NetCDF file.
        decode_times : bool, optional
            If True, decode times encoded in the standard NetCDF datetime format
            into datetime objects. Otherwise, leave them encoded as numbers.
        time_key:
            time variable, defaults to "TIME"

    Returns
    -------
        wf: WaterFrame
    """
    # Create WaterFrame

    time_units = ""
    if decode_times:
        # decode_times in xarray.open_dataset will erase the unit field from TIME, so store it before it is removed
        ds = xr.open_dataset(path, decode_times=False)
        if time_key in ds.variables and "units" in ds[time_units].attrs.keys():
            time_units = ds[time_key].attrs["units"]
        ds.close()

    # Open file with xarray
    ds = xr.open_dataset(path, decode_times=decode_times)

    # Save ds into a WaterFrame
    metadata = dict(ds.attrs)

    df = ds.to_dataframe()

    if time_key in df.columns:
        df = df.set_index(time_key)

    vocabulary = {}
    for variable in ds.variables:
        vocabulary[variable] = dict(ds[variable].attrs)

    if time_units:
        vocabulary[time_key]["units"] = time_units
    return WaterFrame(df, metadata, vocabulary)


def new_read_nc2(filename, decode_times=False):

    import rich
    with nc.Dataset(filename, 'r') as dataset:
        varnames = dataset.variables.keys()
        dimnames = dataset.dimensions.keys()
        rich.print(f"Got {len(dimnames)} dimensions: {dimnames}")
        data = {}
        shapes = []
        manual_sizes = {}
        for var in varnames:
            values = dataset.variables[var][:]
            data[var] = values
            shapes.append(values.shape)
            a = 1
            for s in values.shape:
                a *= s
            manual_sizes[a] = var

        rich.print(manual_sizes)
        key_max = max(manual_sizes.keys())
        max_shape = data[var].shape

        reshaped_data = {}
        for key, value in data.items():
            shape = data[key].shape
            rich.print(f"{key} shape = {shape}")
            reshaped_data[key] = value.reshape(max_shape)

        df = pd.DataFrame(reshaped_data)
        print(df)
    exit()

